{ -------------------------------------------------------------------------}
{ ucalc.pas : Unit include file                                            }
{ Designed to work with Delphi 4.0 or above.                               }
{                                                                          }
{ UCalc Fast Math Parser (Beta N)                                          }
{ Copyright (c) 1998-2005 by Daniel Corbier.  All rights reserved.         }
{ http://www.ucalc.com/mathparser                                          }
{ -------------------------------------------------------------------------}

{$IFNDEF ucAddOnLibrary}
unit ucalc;

interface

{$IFDEF Debug}
uses SysUtils, UCalcFME;
{$ELSE}
uses SysUtils;
//+++ Remove $IF afterwards
{$ENDIF}
{$ENDIF}

type ucDefinitionSpaceType = record
   BaseDefinitionSpace:  Longint;
   DefinitionSpace:      Longint;
   InlineBuffer:         Longint;
   MaxErrorData:         Longint;
   MaxThreadCount:       Longint;
end;

type ucForInternalUseType = record
   LanguageID:           Longint;
   ucOperator:           Pointer;
   ucEvalAddr:           Pointer;
   ucEvaluateAddr:       Pointer;
   ucThreadAddr:         Longword;
   ParamListAddr:        Pointer;
   HandleAddr:           Pointer;
   ReleaseExprPtr:       Pointer;
   TreeIndex:            Longint;
   TreeParamInfoAddress: Pointer;
   HaltEvaluation:       Longint;
   ThreadStatus:         Longint;
   PrevOperationCount:   Longint;
   PrevItemCount:        Longint;
   CompareOld:           Longint;
   IsStringValue:        Longint;
   ucStack:              Longint;
   Infinity:             Longint;
   TrigFactor:          ^Extended;
   TrigFactorInv:       ^Extended;
   DataGroup1:           Longword;
   DataGroup1b:          Longword;
   DataGroup1c:          Longword;
   DataGroup2:           Longword;
   DataGroup2b:          Longword;
   DataGroup2c:          Longword;
   DataGroup3:           Longword;
   DataGroup3b:          Longword;
   DataGroup3c:          Longword;
   DataGroup4:           Longword;
   DataGroup4b:          Longword;
   DataGroup4c:          Longword;
   StringPtr1:           Longword;
end;

type ucPropertyType = record
   DefaultNumericType:          Longint;
   DefaultExternalStringType:   Longint;
   FPU_ControlWord_UCalc:       Longint;
   FPU_ControlWord_HostProgram: Longint;
   FPU_Turbo:                   Longint;
   CaseSensitive:               Longint;
   ComplexNumbers:              Longint;
   ImagSymbol:                  Longint;
   RelationalTrue:              Longint;
   ImplicitMultiply:            Longint;
   MaxParamPerFunction:         Longint;
   ArrayProperties:             Longint;
   FunctionProperties:          Longint;
   StackSize:                   Longint;
   DefaultPrecedence:           Longint;
   OutputDigits:                Longint;
   QuoteChar:                   Longint;
   QuoteCharSynonym:            Longint;
   ArgumentSeparator:           Longint;
   DecimalSeparator:            Longint;
   EscapeChar:                  Longint;
   DefSpace:                    ucDefinitionSpaceType;
end;

type DataByTypeType = record
   case Byte of
      0: (ValuePtr:    Pointer);
      1: (ValueByte:  ^Byte);
      2: (ValueByteS: ^ShortInt);
      3: (ValueInt:   ^SmallInt);
      4: (ValueWrd:   ^Word);
      5: (ValueLng:   ^Longint);
      6: (ValueDwd:   ^Longword);
      7: (ValueQuad:  ^Int64);
      8: (ValueSng:   ^Single);
      9: (ValueDbl:   ^Double);
     10: (ValueExt:   ^Extended);
     11: (ValueCur:   ^Currency);
     12: (ValueStr:   ^AnsiString);
end;

type ucHandleType = record
   ucThread:           Longint;
   Info:               Longword;
   DataType:           Longint;
   DataPtr:            DataByTypeType;
   DataCplxPtr:        DataByTypeType;
   SymbolName:        ^AnsiString;
   SortedHandle:       Longint;
   OverloadHandle:     Longint;
   OpPrecedence:       Longint;
   ReqParamCount:      SmallInt;
   ParamCount:         SmallInt;
   CallbackPtr:        Longint;
   FuncArgs:          ^AnsiString;
   FuncArgHandleData: ^AnsiString;
   FuncArgTypeList:   ^AnsiString;
   VarValueStr:       ^AnsiString;

   DataGroup1:         Longword;
   DataGroup1b:        Longword;
   DataGroup1c:        Longword;
   DataGroup2:         Longword;
   DataGroup2b:        Longword;
   DataGroup2c:        Longword;
   StringPtr1:         Longword;
   StringPtr2:         Longword;
   StringPtr3:         Longword;
   StringPtr4:         Longword;
   StringPtr5:         Longword;

   UserData1:          Longword;
   UserData2:          Longword;
   UserData3:          Longword;
   UserData4:          Longword;
end;

type ucThreadType = record
   ucThread:           Longint;
   Context:            Longint;
   ParamCount:         Longint;
   ErrorNumber:        Longint;
   TreeDirection:      Longint;
   DefSpace:           Longint;
   OperationCount:     Longint;
   ItemCount:          Longint;
   Sequence:           Longint;
   SequenceLast:       Longint;
   ExprHasChanged:     Longint;
   IsFunctionParse:    Longint;
   ComplexStatus:      Longint;
   ucParamAddr:        Pointer;
   ucParamHandleAddr:  Pointer;
   SymbolStart:        Longint;
   SymbolEnd:          Longint;
   PrecedenceMax:      Longint;
   PrecedenceMin:      Longint;
   VarHandle:          Longint;
   OpHandle:           Longint;
   StringWasReturned:  Longint;
   ReturnString:      ^AnsiString;
   ReturnValue:       ^Extended;
   ReturnComplex:     ^Extended;
   ucProperty:         ucPropertyType;
   Handshake:          ucForInternalUseType;
   User1:              Longint;
   User2:              Longint;
   User3:              Longint;
   User4:              Longint;
end;

type ucThreadTypePtr = ^ucThreadType;

type ucComplexNumberType = record
   Real: Extended;
   Imag: Extended;
end;

const
   ucTrue  = 1;
   ucFalse = 0;

   // LanguageID
   uc_Lang_Unsupported    = 0;  
   uc_Lang_Debug          = 1;
   uc_Lang_PowerBasic     = 2;
   uc_Lang_Delphi         = 3;
   uc_Lang_VisualBasic    = 4;
   uc_Lang_VisualBasicNET = 5;
   uc_Lang_CPlusPlus      = 6;

   // Type constants for vType arg of ucDefineFunction and ucDefineVariable
   ucType_Integer8      = 1;
   ucType_Integer8u     = 2;
   ucType_Integer16     = 3;
   ucType_Integer16u    = 4;
   ucType_Integer32     = 5;
   ucType_Integer32u    = 6;
   ucType_Integer64     = 7;
   ucType_Integer64u    = 8;
   ucType_Currency      = 9;
   ucType_CurrencyExt   = 10;
   ucType_FloatSingle   = 11;
   ucType_FloatDouble   = 12;
   ucType_FloatExtended = 13;
   ucType_StringBSTR    = 14;  // For VB
   ucType_StringLPSTR   = 15;  // (ASCIIZ) C++ char default
   ucType_StringAnsi    = 16;  // PowerBASIC and Delphi default
   ucType_Any           = 17;

   ucType_Long          = ucType_Integer32;
   ucType_Double        = ucType_FloatDouble;
   ucType_Extended      = ucType_FloatExtended;
   ucType_FixedPt       = ucType_Currency;
   ucType_Quad          = ucType_Integer64u;
   ucType_AnsiString    = ucType_StringAnsi;

   // Context options (t.Context)
   ucContext_Escape            = 1;
   ucContext_Try               = 2;
   ucContext_ucDefine          = 4;
   ucContext_ucDefineFunction  = 8;
   ucContext_ucDefineVariable  = 16;
   ucContext_ucEval            = 32;
   ucContext_ucEvalStr         = 64;
   ucContext_ucEvaluateStr     = 128;
   ucContext_ucParse           = 256;
   ucContext_ucParseOrucEval   = 512;
   ucContext_ucReset           = 1024;
   ucContext_Reserved          = 2048;

   // Options for fProperties parameter of ucDefineFunction
   // and for ucProperty.FunctionProperties
   ucLeftToRight  = 0;
   ucInline       = 1;
   ucDynamic      = 2;
   ucFastCallback = 4;
   ucLocal        = 8;
   ucComplex      = 16;
   ucLocked       = 32;
   ucConstant     = ucLocked;
   ucOverride     = 64;
   ucOverwrite    = 128;
   ucOverload     = 256;
   ucRecursion    = 512;
   ucBootstrap    = 1024;
   ucDelaySyntax  = 1024;
   ucRightToLeft  = 2048;

   // For ucTreeProperty, ucTreeNodeProperty,
   // ucTreeParamProperty, and ucTreeParentProperty
   ucIsInline            = 0;
   ucIsDynamic           = 1;
   ucIsFastCallback      = 2;
   ucIsLocal             = 3;
   ucIsComplex           = 4;
   ucIsLocked            = 5;
   ucIsConstant          = ucIsLocked;
   ucIsOverride          = 6;
   ucIsOverwrite         = 7;
   ucIsOverload          = 8;
   ucIsRecursion         = 9;
   ucIsBootstrap         = 10;
   ucIsDelaySyntax       = 10;
   ucIsRightToLeft       = 11;

   ucIsArray             = 12;

   ucIsFunction          = 13;
   ucIsOperator          = 14;
   ucIsInfixOp           = 15;
   ucIsPostfixOp         = 16;
   ucIsPrefixOp          = 17;
   ucIsVariable          = 18;
   ucIsJITVariable       = 19;
   ucIsJITExpression     = 20;
   ucIsString            = 21;
   ucIsExternal          = 22;
   ucIsActive            = 23;
   ucIsItemCount         = 24;
   ucIsSorted            = 25;
   ucIsMaybeComplex      = 26;
   ucIsComplexAltDef     = 27;
   ucIsArgTypeOverload   = 28;
   ucIsSyntax            = 29;
   Reserved1             = 30;
   Reserved2             = 31;

   ucInfixOp   = ucIsInfixOp;
   ucPostfixOp = ucIsPostfixOp;
   ucPrefixOp  = ucIsPrefixOp;

// For ucElementCategory
   ucIsOperand   = 32;
// ucIsFunction  = (already defined)
// ucIsInfixOp   = (already defined)
// ucIsPrefixOp  = (already defined)
// ucIsPostfixOp = (already defined)

   ucDataType    = 40;
   ucResultPtr   = 41;
   ucSignPos     = 42;
   ucElementCat  = 43;
   ucTreeIndex   = 44;
   ucErrorNumber = 45;
   ucOpHandle    = 46;
   ucFuncHandle  = ucOpHandle;
   ucExprPtr     = 47;
   ucCodeAddress = 48;
   ucCurrentPtr  = 49;
   ucParentPtr   = 50;
   ucDataHandle  = 51;
   ucMiscData    = 52;
   ucUserData1   = 53;
   ucUserData2   = 54;
   ucUserData3   = 55;
   ucUserData4   = 56;

   ucTreePre       = 5;
   ucTreeDownEarly = 6;
   ucTreeDown      = 7;
   ucTreeUp        = 8;

   ucErrorCallback = 9;

   ucNoChange    = 0;
   ucMinorChange = 1;
   ucMajorChange = 2;

   // Trig mode options for ucSetTrigMode
   ucRadians = 1;
   ucDegrees = 2;
   ucGrads   = 3;
   ucCycles  = 4;

   // Options for ucArrayInfo,
   // and the first two for ucProperty.ArrayProperties
   ucArray_BoundsCheck    = 1;
   ucArray_ColumnMajor    = 2;
   ucArray_DataAddress    = 3;
   ucArray_DataType       = 4;
   ucArray_DimensionCount = 5;
   ucArray_ElementCount   = 6;
   ucArray_ElementSize    = 7;

   // FPU control word values for ucProperty.FPU_ControlWord,
   // and SetFPU(), SetOrFPU, SetXorFPU, GetFPU(), RestoreFPU
   ucFPU_ErrMask_InvalidOp      = 1;
   ucFPU_ErrMask_DenormalOp     = 2;
   ucFPU_ErrMask_ZeroDivide     = 4;
   ucFPU_ErrMask_Overflow       = 8;
   ucFPU_ErrMask_Underflow      = 16;
   ucFPU_ErrMask_PrecisionLoss  = 32;

   ucFPU_Precision_Single       = 0;
   ucFPU_Precision_Double       = 512;
   ucFPU_Precision_Extended     = 768;

   ucFPU_Round_Even             = 0;
   ucFPU_Round_Down             = 1024;
   ucFPU_Round_Up               = 2048;
   ucFPU_Round_Truncate         = 3072;

   // For ucProperty.FPU_Turbo
   ucFPU_Safe      = 0;
   ucFPU_SkipSet   = 1;
   ucFPU_SkipReset = 65536;

   // For ucSetErrorData / ucGetErrorData
   ucNewIndex = -2;
   ucNewUniqueIndex  = -1;
   ucAllErrorData    = -1;

   // Error Values
   ucErr_None                = 0;
   ucErr_ucErrorData         = 1;
   ucErr_SyntaxError         = 2;
   ucErr_MismatchedParenth   = 3;
   ucErr_MismatchedQuotes    = 4;
   ucErr_UndefinedVariable   = 5;
   ucErr_UndefinedFunction   = 6;
   ucErr_IncorrectArgType    = 7;
   ucErr_InvalidNumberOfArgs = 8;
   ucErr_InvalidOperandType  = 9;
   ucErr_InvalidVariableName = 10;
   ucErr_ImplicitMult        = 11;
   ucErr_IncompleteBlock     = 12;

   ucErr_InvalidOperation    = 20;
   ucErr_DenormalOperand     = 21;
   ucErr_DivisionByZero      = 22;
   ucErr_Overflow            = 23;
   ucErr_Underflow           = 24;
   ucErr_InexactResult       = 25;
   ucErr_SubscriptOutOfRange = 35;
   ucErr_StackOverflow       = 36;

   ucErr_OperationDenied     = 40;
   ucErr_PropertyConflict    = 41;
   ucErr_ConstantWriteDenied = 42;
   ucErr_FunctionDefError    = 43;
   ucErr_FunctionLocked      = 44;

   ucErr_DefinitionSpaceFull = 60;
   ucErr_ThreadInitFailed    = 61;
   ucErr_UnsupportedFeature  = 62;

   Op_Func_GetFPU     = 36;
   Op_Func_SetFPU     = 37;
   Op_Func_SetOrFPU   = 38;
   Op_Func_SetXorFPU  = 39;
   Op_Func_RestoreFPU = 40;
   Op_Func_Ceil       = 41;
   Op_Func_Factorial  = 42;
   Op_Func_Not        = 43;
   Op_Cur_Add         = 44;
   Op_Cur_Multiply    = 45;
   Op_Cur_Subtract    = 46;
   Op_Cur_DivideToExt = 47;
   Op_Qwd_Add         = 48;
   Op_Qwd_Multiply    = 49;
   Op_Qwd_Subtract    = 50;
   Op_Qwd_DivideToInt = 51;
   Op_Qwd_DivideToExt = 52;
   Op_Int_Add         = 53;
   Op_Int_Multiply    = 54;
   Op_Int_Subtract    = 55;
   Op_Int_DivideToInt = 56;
   Op_Int_DivideToExt = 57;
   Op_Int_Mod         = 58;
   Op_IntPower        = 59;

var
   t0: ucThreadType;
   ucItem: array of ucHandleType;

{$IFNDEF ucAddOnLibrary}
// The following are defined outside of the class
// They are for callback functions (except for ucReset, which has a global effect)
function ucElementCategory(var t: ucThreadType): Longint;
function ucError(): Longint;
function ucGetErrorMessage(ErrorNumber: Longint = -1; t: Longint = 0): PChar;
function ucGetExpr(var t: ucThreadType): PChar;
function ucTreeGetExpr(var t: ucThreadType): PChar;
function ucGetParamStr(ParamIndex: Longint; var t: ucThreadType): PChar;
function ucGetErrorData(errIndex: Longint; t: Longint = 0): PChar;
function ucNewSymbolHandle(var t: ucThreadType): Longint;
function ucParam(x: Longint; var t: ucThreadType): Extended;
function ucParamCplx(x: Longint; var t: ucThreadType): Extended;
function ucParamStr(ParamIndex: Longint; var t: ucThreadType): PChar;
function ucParamHandle(pIndex: Longint; var t: ucThreadType): Longint;
function ucParamType(pIndex: Longint; var t: ucThreadType): Longint;
function ucTreeGetFunc(var t: ucThreadType): PChar;
function ucTreeGetOp(var t: ucThreadType): PChar;
function ucTreeGetParam(pIndex: Longint; var t: ucThreadType): PChar;
function ucTreeGetParamList(var t: ucThreadType): PChar;
function ucTreeProperty(NodePtr: Longword; pIndex, pOption: Longint; var t: ucThreadType): Longint;
function ucTreeNodeProperty(pOption: Longint; var t: ucThreadType): Longint;
function ucTreeParamProperty(pIndex, pOption: Longint; var t: ucThreadType): Longint;
function ucTreeParentProperty(pOption: Longint; var t: ucThreadType): Longint;
procedure ucReset(var t: ucThreadType);
procedure ucReturn(vValue: Extended; var t: ucThreadType);
procedure ucReturnCplx(vValue: Extended; var t: ucThreadType);
procedure ucReturnStr(Expression: AnsiString; var t: ucThreadType);
procedure ucSetErrorData(errIndex: Longint; vValue: AnsiString; var t: ucThreadType);
procedure ucSetExpr(Expression: AnsiString; var t: ucThreadType);
procedure ucSetParamStr(ParamIndex: Longint; ArgStr: AnsiString; var t: ucThreadType);
procedure ucTreeSetExpr(Expression: AnsiString; var t: ucThreadType);
procedure ucTreeSetFunc(TreeFunc: AnsiString; var t: ucThreadType);
procedure ucTreeSetOp(TreeOp: AnsiString; var t: ucThreadType);
procedure ucTreeSetParam(pIndex: Longint; pData: AnsiString; var t: ucThreadType);
procedure ucTreeSetParamList(pData: AnsiString; var t: ucThreadType);

//  The following functions are for single-thread usage ( [[+++ not all of them]])
function ucAlias(AssignFrom, AssignTo: AnsiString; aProperties: Longint = 0; Precedence: Longint = 0): Longint;
function ucArrayInfo(VarHandle, InfoOption: Longint): Longword;
function ucDefineConstant(ConstantDef: AnsiString; vProperties: Longint = 0): Longint;
function ucDefine(Definition: AnsiString): Longint;
function ucDefineFunction(FunctionDef: AnsiString; FunctionAddress: Pointer = nil; fProperties: Longint = 0; vType: Longint = 0): Longint;
function ucDefineMacro(Definition: AnsiString): Longint;
function ucDefineOperator(OperatorDef: AnsiString; Precedence: Longint = 0; OperatorAddress: Pointer = nil; fProperties: Longint = 0; vType: Longint = 0): Longint;
function ucDefineSyntax(Definition: AnsiString): Longint;
function ucDefineVariable(VariableDef: AnsiString; vProperties: Longint = 0): Longint;
function ucAttachVariable(VariableDef: AnsiString; vPtr: Pointer; ExprPtr: Pointer = nil; vProperties: Longint = 0; t: Longint = 0): Longint;
function ucErrorData(errIndex: Longint; t: Longint = 0): PChar;
function ucErrorHandler(FunctionAddress: Pointer): Longint;
function ucErrorMessage(ErrorNumber: Longint = -1; t: Longint = 0): PChar;
function ucEval(Expression: AnsiString; t: Longint = 0): Extended;
function ucEvalStr(Expression: AnsiString; t: Longint = 0): PChar;
function ucEvaluate(ExpressionPtr: Pointer): Extended; overload;
function ucEvaluate(ExpressionPtr: Pointer; var ucValue: ucComplexNumberType; t: Longint = 0): Extended; overload;
function ucEvaluateCplx(ExpressionPtr: Pointer; var ucValue: ucComplexNumberType; t: Longint = 0): Extended;
function ucEvaluateStr(ExpressionPtr: Pointer): PChar;
function ucGetThreadInfo(ucThread: Longint; var t: ucThreadType): Pointer;
function ucGetVariableValue(VarHandle: Longint): Extended;
function ucGetVariableValueStr(VarHandle: Longint): PChar;
function ucImagPart(t: Longint = 0): Extended;
function ucLowerBound(VarHandle: Longint; Dimension: Longint = 1): Longint;
function ucParse(Expression: AnsiString; LocalVars: AnsiString = ''; ExprType: Longint = 0): Pointer;
function ucPrecedence(Op: AnsiString; OpCategory: Longint = 0; t: Longint = 0): Longint;
function ucPreParse(FunctionAddress: Pointer): Longint;
function ucSymbolHandle(ucSymbol: AnsiString; SymbolProperty: Longint = 0; t: Longint = 0): Longint;
function ucThreadAddr(t: Longint): ucThreadTypePtr;
function ucTreeParseDownEarly(FunctionAddress: Pointer): Longint;
function ucTreeParseDown(FunctionAddress: Pointer): Longint;
function ucTreeParseUp(FunctionAddress: Pointer): Longint;
function ucUpperBound(VarHandle: Longint; Dimension: Longint = 1): Longint;
function ucVariableValueStr(VarHandle: Longint): PChar;
procedure ucDetachSymbol(Symbol: AnsiString; t: Longint = 0);
procedure ucErrorClear(var t: ucThreadType);
procedure ucLicense(SerialNumber: AnsiString);
procedure ucReleaseExprPtr(ExpressionPtr: Pointer);
procedure ucReleaseItem(ItemHandle: Longint);
procedure ucSetThreadAddr(ucThread: Longint; var t: ucThreadType);
procedure ucSetTrigMode(TrigMode: Longint; t: Longint = 0);
procedure ucSetVariableValue(VarHandle: Longint; vValue: Extended);
procedure ucSetVariableValueStr(VarHandle: Longint; vValue: AnsiString);

type FastMath = class
private
   t: Longint;

   procedure ucReleaseThread(ucThread: Longint);
   function ucNewThread: Longint;

   protected
   published

public
   constructor Create;
   destructor Destroy; override;

   function ucAlias(AssignFrom, AssignTo: AnsiString; aProperties: Longint = 0; Precedence: Longint = 0): Longint;
   function ucAttachVariable(VariableDef: AnsiString; vPtr: Pointer; ExprPtr: Pointer = nil; vProperties: Longint = 0): Longint;
   function ucDefineConstant(ConstantDef: AnsiString; vProperties: Longint = 0): Longint;
   function ucDefine(Definition: AnsiString): Longint;
   function ucDefineFunction(FunctionDef: AnsiString; FunctionAddress: Pointer = nil; fProperties: Longint = 0; vType: Longint = 0): Longint;
   function ucDefineMacro(Definition: AnsiString): Longint;
   function ucDefineOperator(OperatorDef: AnsiString; Precedence: Longint = 0; OperatorAddress: Pointer = nil; fProperties: Longint = 0; vType: Longint = 0): Longint;
   function ucDefineSyntax(Definition: AnsiString): Longint;
   function ucDefineVariable(VariableDef: AnsiString; vProperties: Longint = 0): Longint;
   function ucError(): Longint;
   function ucErrorData(errIndex: Longint): PChar;
   function ucErrorHandler(FunctionAddress: Pointer): Longint;
   function ucErrorMessage(ErrorNumber: Longint = -1): PChar;
   function ucEval(Expression: AnsiString): Extended;
   function ucEvalStr(Expression: AnsiString): PChar;
   function ucEvaluate(ExpressionPtr: Pointer): Extended; overload;
   function ucEvaluate(ExpressionPtr: Pointer; var ucValue: ucComplexNumberType): Extended; overload;
   function ucEvaluateCplx(ExpressionPtr: Pointer; var ucValue: ucComplexNumberType): Extended;
   function ucEvaluateStr(ExpressionPtr: Pointer): PChar;
   function ucGetErrorData(errIndex: Longint): PChar;
   function ucGetThreadIndex: Longint;
   function ucGetVariableValue(VarHandle: Longint): Extended;
   function ucGetVariableValueStr(VarHandle: Longint): PChar;
   function ucImagPart: Extended;
   function ucParse(Expression: AnsiString; LocalVars: AnsiString = ''; ExprType: Longint = 0): Pointer;
   function ucPrecedence(Op: AnsiString; OpCategory: Longint = 0): Longint;
   function ucPreParse(FunctionAddress: Pointer): Longint;
   function ucSymbolHandle(ucSymbol: AnsiString; SymbolProperty: Longint = 0): Longint;
   function ucTreeParseDownEarly(FunctionAddress: Pointer): Longint;
   function ucTreeParseDown(FunctionAddress: Pointer): Longint;
   function ucTreeParseUp(FunctionAddress: Pointer): Longint;
   function ucVariableValueStr(VarHandle: Longint): PChar;
   procedure ucDetachSymbol(Symbol: AnsiString);
   procedure ucErrorClear;
   procedure ucReleaseItem(ItemHandle: Longint);
   procedure ucReleaseExprPtr(ExpressionPtr: Pointer);
   procedure ucSetThreadIndex(ucThread: Longint);
   procedure ucSetTrigMode(TrigMode: Longint);
   procedure ucSetVariableValue(VarHandle: Longint; vValue: Extended);
   procedure ucSetVariableValueStr(VarHandle: Longint; vValue: AnsiString);
end;


implementation
{$ENDIF}

function ucAliasE(AssignFrom, AssignTo: AnsiString; aProperties, Precedence: Longint; var t: ucThreadType): Longint; stdcall; external 'ucalcfmp.dll';
function ucArrayInfoE(VarHandle, InfoOption: Longint): Longword; stdcall; external 'ucalcfmp.dll';
function ucDefineE(Definition: AnsiString; var t: ucThreadType): Longint; stdcall; external 'ucalcfmp.dll';
function ucDefineFunctionE(FunctionDef: AnsiString; FunctionAddress: Pointer; Precedence, fProperties, vType: Longint; var t: ucThreadType): Longint; stdcall; external 'ucalcfmp.dll';
function ucDefineVariableE(VariableDef: AnsiString; vPtr, Expr: Pointer; vProperties, vType: Longint; var t: ucThreadType): Longint; stdcall; external 'ucalcfmp.dll';
function ucEvalE(Expression: AnsiString; var t: ucThreadType): Extended; stdcall; external 'ucalcfmp.dll';
function ucEvalStrE(Expression: AnsiString; var t: ucThreadType): PChar; stdcall; external 'ucalcfmp.dll';
function ucEvaluateStrE(ExpressionPtr: Pointer): PChar; stdcall; external 'ucalcfmp.dll';
function ucGetErrorDataE(errIndex: Longint; var t: ucThreadType): PChar; stdcall; external 'ucalcfmp.dll';
function ucGetParamE(pIndex: Longint; var t: ucThreadType): Extended; stdcall; external 'ucalcfmp.dll';
function ucGetParamCplxE(pIndex: Longint; var t: ucThreadType): Extended; stdcall; external 'ucalcfmp.dll';
function ucGetParamStrE(ParamIndex: Longint; var t: ucThreadType): PChar; stdcall; external 'ucalcfmp.dll';
function ucGetThreadInfoE(ucThread: Longint; var t: ucThreadType): Pointer; stdcall; external 'ucalcfmp.dll';
function ucGetVariableValueE(VarHandle: Longint): Extended; stdcall; external 'ucalcfmp.dll';
function ucGetVariableValueStrE(VarHandle: Longint): PChar; stdcall; external 'ucalcfmp.dll';
function ucImagPartE(var t: ucThreadType): Extended; stdcall; external 'ucalcfmp.dll';
function ucLowerBoundE(VarHandle, Dimension: Longint): Longint; stdcall; external 'ucalcfmp.dll';
function ucNewThreadE: Longint; stdcall; external 'ucalcfmp.dll';
function ucNewSymbolHandleE(var t: ucThreadType): Longint; stdcall; external 'ucalcfmp.dll';
function ucOpE(OpID: Longint): Pointer; stdcall; external 'ucalcfmp.dll';
function ucParamHandleE(pIndex: Longint; var t: ucThreadType): Longint; stdcall; external 'ucalcfmp.dll';
function ucParamTypeE(pIndex: Longint; var t: ucThreadType): Longint; stdcall; external 'ucalcfmp.dll';
function ucParseE(Expression, LocalVars: AnsiString; ExprType: Longint; var t: ucThreadType): Pointer; stdcall; external 'ucalcfmp.dll';
function ucPrecedenceE(Op: AnsiString; OpCategory: Longint; var t: ucThreadType): Longint; stdcall; external 'ucalcfmp.dll';
function ucSetCallbackE(FunctionAddress: Pointer; TreeOption: Longint; var t: ucThreadType): Longint; stdcall; external 'ucalcfmp.dll';
function ucSymbolHandleE(ucSymbol: AnsiString; SymbolProperty: Longint; var t: ucThreadType): Longint; stdcall; external 'ucalcfmp.dll';
function ucThreadAddrE(t: Longint): ucThreadTypePtr; stdcall; external 'ucalcfmp.dll';
function ucTreeGetParamE(pIndex: Longint; var t: ucThreadType): PChar; stdcall; external 'ucalcfmp.dll';
function ucTreePropertyE(NodePtr: Longword; pIndex, pOption: Longint; var t: ucThreadType): Longint; stdcall; external 'ucalcfmp.dll';
function ucTreeSetParamE(pIndex: Longint; pData: AnsiString; var t: ucThreadType): Longint; stdcall; external 'ucalcfmp.dll';
function ucUpperBoundE(VarHandle, Dimension: Longint): Longint; stdcall; external 'ucalcfmp.dll';
procedure ucDetachSymbolE(Symbol: AnsiString; var t: ucThreadType); stdcall; external 'ucalcfmp.dll';
procedure ucErrorClearE(var t: ucThreadType); stdcall; external 'ucalcfmp.dll';
procedure ucLicenseE(SerialNumber: AnsiString); stdcall; external 'ucalcfmp.dll';
procedure ucReleaseItemE(ItemHandle: Longint); stdcall; external 'ucalcfmp.dll';
procedure ucReleaseThreadE(ucThread: Longint); stdcall; external 'ucalcfmp.dll';
procedure ucResetE(var t: ucThreadType); stdcall; external 'ucalcfmp.dll';
procedure ucReturnE(vValue: Extended; var t: ucThreadType); stdcall; external 'ucalcfmp.dll';
procedure ucReturnCplxE(vValue: Extended; var t: ucThreadType); stdcall; external 'ucalcfmp.dll';
procedure ucReturnStrE(Expression: AnsiString; var t: ucThreadType); stdcall; external 'ucalcfmp.dll';
procedure ucSetErrorDataE(errIndex: Longint; vValue: AnsiString; var t: ucThreadType); stdcall; external 'ucalcfmp.dll';
procedure ucSetParamStrE(ParamIndex: Longint; Argument: AnsiString; var t: ucThreadType); stdcall; external 'ucalcfmp.dll';
procedure ucSetThreadAddrE(ucThread: Longint; var t: ucThreadType); stdcall; external 'ucalcfmp.dll';
procedure ucSetTrigModeE(TrigMode: Longint; var t: ucThreadType); stdcall; external 'ucalcfmp.dll';
procedure ucSetVariableValueE(VarHandle: Longint; vValue: Extended); stdcall; external 'ucalcfmp.dll';
procedure ucSetVariableValueStrE(VarHandle: Longint; vValue: AnsiString); stdcall; external 'ucalcfmp.dll';

// These 3 functions come from ucalcfmE.dll instead of ucalcfmP.dll
{$IFNDEF Debug}
// +++ Remove $IF afterwards
function ucEvaluateE(ExpressionPtr: Pointer): Extended; stdcall; external 'ucalcfme.dll';
function ucEvaluateLngE(ExpressionPtr: Pointer): Longint; stdcall; external 'ucalcfme.dll';
function ucEvaluateQwdE(ExpressionPtr: Pointer): Int64; stdcall; external 'ucalcfme.dll';
function ucEvaluateFxtE(ExpressionPtr: Pointer): Currency; stdcall; external 'ucalcfme.dll';
procedure ucEvaluateInitE(); stdcall; external 'ucalcfme.dll';
procedure ucReleaseExprPtrE(ExpressionPtr: Pointer); stdcall; external 'ucalcfme.dll';
{$ENDIF}

function ucThreadAddr(t: Longint): ucThreadTypePtr;
begin
   Result := ucThreadAddrE(t);
end;

function ucElementCategory(var t: ucThreadType): Longint;
begin
   Result := ucTreePropertyE(0, 0, ucElementCat, t);
end;

function ucTreeProperty(NodePtr: Longword; pIndex, pOption: Longint; var t: ucThreadType): Longint;
begin
   Result := ucTreePropertyE(NodePtr, pIndex, pOption, t);
end;

function ucTreeNodeProperty(pOption: Longint; var t: ucThreadType): Longint;
begin
   Result := ucTreePropertyE(0, 0, pOption, t);
end;

function ucTreeParamProperty(pIndex, pOption: Longint; var t: ucThreadType): Longint;
begin
   Result := ucTreePropertyE(0, pIndex, pOption, t);
end;

function ucTreeParentProperty(pOption: Longint; var t: ucThreadType): Longint;
var
   ParentPtr: Longword;
begin
   ParentPtr := ucTreePropertyE(0, 0, ucParentPtr, t);
   Result := ucTreePropertyE(ParentPtr, 0, pOption, t);
end;

function ucNewSymbolHandle(var t: ucThreadType): Longint;
begin
   Result := ucNewSymbolHandleE(t);
end;

function ucParamHandle(pIndex: Longint; var t: ucThreadType): Longint;
begin
   Result := ucParamHandleE(pIndex, t);
end;

function ucParamType(pIndex: Longint; var t: ucThreadType): Longint;
begin
   Result := ucParamTypeE(pIndex, t);
end;

function ucParam(x: Longint; var t: ucThreadType): Extended;
begin
   // +++ use array instead of ucGetParamE (or place ucGetParamE in ucalcfme.dll)
   Result := ucGetParamE(x, t);
end;

function ucParamCplx(x: Longint; var t: ucThreadType): Extended;
begin
   // +++ use array instead of ucGetParamE (or place ucGetParamE in ucalcfme.dll)
   Result := ucGetParamCplxE(x, t);
end;

function ucGetExpr(var t: ucThreadType): PChar;
begin
   Result := ucTreeGetParamE(-1, t);
end;

function ucTreeGetExpr(var t: ucThreadType): PChar;
begin
   Result := ucTreeGetParamE(-1, t);
end;

function ucTreeGetFunc(var t: ucThreadType): PChar;
begin
   Result := ucTreeGetParamE(0, t);
end;

function ucTreeGetOp(var t: ucThreadType): PChar;
begin
   Result := ucTreeGetParamE(0, t);
end;

function ucTreeGetParam(pIndex: Longint; var t: ucThreadType): PChar;
begin
   Result := ucTreeGetParamE(pIndex, t);
end;

function ucTreeGetParamList(var t: ucThreadType): PChar;
begin
   ucTreeGetParamList := ucTreeGetParamE(-2, t);
end;

procedure ucSetExpr(Expression: AnsiString; var t: ucThreadType);
begin
   ucTreeSetParamE(-1, Expression, t);
end;

procedure ucTreeSetExpr(Expression: AnsiString; var t: ucThreadType);
begin
   ucTreeSetParamE(-1, Expression, t);
end;

procedure ucTreeSetFunc(TreeFunc: AnsiString; var t: ucThreadType);
begin
   ucTreeSetParamE(0, TreeFunc, t);
end;

procedure ucTreeSetOp(TreeOp: AnsiString; var t: ucThreadType);
begin
   ucTreeSetParamE(0, TreeOp, t);
end;

procedure ucTreeSetParam(pIndex: Longint; pData: AnsiString; var t: ucThreadType);
begin
   ucTreeSetParamE(pIndex, pData, t);
end;

procedure ucTreeSetParamList(pData: AnsiString; var t: ucThreadType);
begin
   ucTreeSetParamE(-2, pData, t);
end;

function ucErrorHandler(FunctionAddress: Pointer): Longint;
begin
   Result := ucSetCallbackE(FunctionAddress, ucErrorCallback, ucThreadAddr(0)^);
end;

function ucGetParamStr(ParamIndex: Longint; var t: ucThreadType): PChar;
begin
   Result := ucGetParamStrE(ParamIndex, t);
end;

function ucParamStr(ParamIndex: Longint; var t: ucThreadType): PChar;
begin
   Result := ucGetParamStrE(ParamIndex, t);
end;

procedure ucLicense(SerialNumber: AnsiString);
begin
   ucLicenseE(SerialNumber);
end;

procedure ucSetParamStr(ParamIndex: Longint; ArgStr: AnsiString; var t: ucThreadType);
begin
   ucSetParamStrE(ParamIndex, ArgStr, t);
end;

procedure ucReturn(vValue: Extended; var t: ucThreadType);
begin
   ucReturnE(vValue, t);
end;

procedure ucReturnCplx(vValue: Extended; var t: ucThreadType);
begin
   ucReturnCplxE(vValue, t);
end;

procedure ucReturnStr(Expression: AnsiString; var t: ucThreadType);
begin
   ucReturnStrE(Expression, t);
end;

procedure ucSetThreadAddr(ucThread: Longint; var t: ucThreadType);
begin
   ucSetThreadAddrE(ucThread, t);
end;

procedure ucSetErrorData(errIndex: Longint; vValue: AnsiString; var t: ucThreadType);
begin
   ucSetErrorDataE(errIndex, vValue, t);
end;


//  These function definitions are for single-thread and/or startup usage only [[+++ Revise this statement]]
procedure ucDetachSymbol(Symbol: AnsiString; t: Longint = 0);
begin
   ucDetachSymbolE(Symbol, ucThreadAddr(t)^);
end;

procedure ucErrorClear(var t: ucThreadType);
begin
   ucErrorClearE(t);
end;

procedure ucReleaseItem(ItemHandle: Longint);
begin
   ucReleaseItemE(ItemHandle);
end;

procedure ucReleaseExprPtr(ExpressionPtr: Pointer);
begin
   ucReleaseExprPtrE(ExpressionPtr);
end;

procedure ucSetTrigMode(TrigMode: Longint; t: Longint = 0);
begin
   ucSetTrigModeE(TrigMode, ucThreadAddr(t)^);
end;

procedure ucSetVariableValue(VarHandle: Longint; vValue: Extended);
begin
   ucSetVariableValueE(VarHandle, vValue);
end;

procedure ucSetVariableValueStr(VarHandle: Longint; vValue: AnsiString);
begin
   ucSetVariableValueStrE(VarHandle, vValue);
end;

function ucAlias(AssignFrom, AssignTo: AnsiString; aProperties: Longint = 0; Precedence: Longint = 0): Longint;
begin
   Result := ucAliasE(AssignFrom, AssignTo, aProperties, Precedence, ucThreadAddr(0)^);
end;

function ucArrayInfo(VarHandle, InfoOption: Longint): Longword;
begin
   Result := ucArrayInfoE(VarHandle, InfoOption);
end;

function ucDefine(Definition: AnsiString): Longint;
begin
   Result := ucDefineE(Definition, ucThreadAddr(0)^);
end;

function ucDefineConstant(ConstantDef: AnsiString; vProperties: Longint = 0): Longint;
begin
   Result := ucDefineVariableE(ConstantDef, nil, nil, vProperties+ucConstant, t0.ucProperty.DefaultNumericType, ucThreadAddr(0)^);
end;

function ucDefineFunction(FunctionDef: AnsiString; FunctionAddress: Pointer = nil; fProperties: Longint = 0; vType: Longint = 0): Longint;
begin
   Result := ucDefineFunctionE(FunctionDef, FunctionAddress, 0, fProperties, vType, ucThreadAddr(0)^);
end;

function ucDefineMacro(Definition: AnsiString): Longint;
begin
   Result := ucDefineE('Macro '+Definition, ucThreadAddr(0)^);
end;

function ucDefineOperator(OperatorDef: AnsiString; Precedence: Longint = 0; OperatorAddress: Pointer = nil; fProperties: Longint = 0; vType: Longint = 0): Longint;
begin
   Result := ucDefineFunctionE(OperatorDef, OperatorAddress, Precedence, fProperties, vType, ucThreadAddr(0)^);
end;

function ucDefineSyntax(Definition: AnsiString): Longint;
begin
   Result := ucDefineE('Syntax '+Definition, ucThreadAddr(0)^);
end;

function ucDefineVariable(VariableDef: AnsiString; vProperties: Longint = 0): Longint;
begin
   Result := ucDefineVariableE(VariableDef, nil, nil, vProperties, t0.ucProperty.DefaultNumericType, ucThreadAddr(0)^);
end;

function ucAttachVariable(VariableDef: AnsiString; vPtr: Pointer; ExprPtr: Pointer = nil; vProperties: Longint = 0; t: Longint = 0): Longint;
begin
   Result := ucDefineVariableE(VariableDef, vPtr, ExprPtr, vProperties, t0.ucProperty.DefaultNumericType, ucThreadAddr(0)^);
end;

function ucEval(Expression: AnsiString; t: Longint = 0): Extended;
begin
   Result := ucEvalE(Expression, ucThreadAddr(t)^);
end;

function ucEvalStr(Expression: AnsiString; t: Longint = 0): PChar;
begin
   Result := ucEvalStrE(Expression, ucThreadAddr(t)^);
end;

function ucEvaluate(ExpressionPtr: Pointer): Extended; overload;
begin
   Result := ucEvaluateE(ExpressionPtr);
end;

function ucEvaluateLng(ExpressionPtr: Pointer): Longint; overload;
begin
   Result := ucEvaluateLngE(ExpressionPtr);
end;

function ucEvaluateQwd(ExpressionPtr: Pointer): Int64; overload;
begin
   Result := ucEvaluateQwdE(ExpressionPtr);
end;

function ucEvaluateFxt(ExpressionPtr: Pointer): Currency; overload;
begin
   Result := ucEvaluateFxtE(ExpressionPtr);
end;

function ucEvaluate(ExpressionPtr: Pointer; var ucValue: ucComplexNumberType; t: Longint = 0): Extended; overload;
begin
   Result := ucEvaluateE(ExpressionPtr);
   ucValue.Real := Result;
   ucValue.Imag := ucImagPartE(ucThreadAddr(t)^);
end;

function ucEvaluateCplx(ExpressionPtr: Pointer; var ucValue: ucComplexNumberType; t: Longint = 0): Extended;
begin
   Result := ucEvaluateE(ExpressionPtr);
   ucValue.Real := Result;
   ucValue.Imag := ucImagPartE(ucThreadAddr(t)^);
end;

function ucEvaluateStr(ExpressionPtr: Pointer): PChar;
begin
   Result := ucEvaluateStrE(ExpressionPtr);
end;

function ucError(): Longint;
begin
   Result := t0.ErrorNumber;
end;

function ucErrorData(errIndex: Longint; t: Longint = 0): PChar;
begin
   Result := ucGetErrorDataE(errIndex, ucThreadAddr(t)^);
end;

function ucGetErrorData(errIndex: Longint; t: Longint = 0): PChar;
begin
   Result := ucGetErrorDataE(errIndex, ucThreadAddr(t)^);
end;

function ucGetThreadInfo(ucThread: Longint; var t: ucThreadType): Pointer;
begin
   Result := ucGetThreadInfoE(ucThread, t);
end;

function ucGetVariableValue(VarHandle: Longint): Extended;
begin
   Result := ucGetVariableValueE(VarHandle);
end;

function ucGetVariableValueStr(VarHandle: Longint): PChar;
begin
   Result := ucGetVariableValueStrE(VarHandle);
end;

function ucImagPart(t: Longint = 0): Extended;
begin
   Result := ucImagPartE(ucThreadAddr(t)^);
end;

function ucLowerBound(VarHandle: Longint; Dimension: Longint = 1): Longint;
begin
   Result := ucLowerBoundE(VarHandle, Dimension);
end;

function ucOp(OpID: Longword): Pointer;
begin
   Result := ucOpE(OpID);
end;

function ucParse(Expression: AnsiString; LocalVars: AnsiString = ''; ExprType: Longint = 0): Pointer;
begin
   Result := ucParseE(Expression, LocalVars, ExprType, ucThreadAddr(0)^);
end;

function ucPrecedence(Op: AnsiString; OpCategory: Longint = 0; t: Longint = 0): Longint;
begin
  Result := ucPrecedenceE(Op, OpCategory, ucThreadAddr(t)^)
end;

function ucPreParse(FunctionAddress: Pointer): Longint;
begin
   Result := ucSetCallbackE(FunctionAddress, ucTreePre, ucThreadAddr(0)^);
end;

function ucSymbolHandle(ucSymbol: AnsiString; SymbolProperty: Longint = 0; t: Longint = 0): Longint;
begin
   Result := ucSymbolHandleE(ucSymbol, SymbolProperty, ucThreadAddr(t)^);
end;

function ucTreeParseDownEarly(FunctionAddress: Pointer): Longint;
begin
   Result := ucSetCallbackE(FunctionAddress, ucTreeDownEarly, ucThreadAddr(0)^);
end;

function ucTreeParseDown(FunctionAddress: Pointer): Longint;
begin
   Result := ucSetCallbackE(FunctionAddress, ucTreeDown, ucThreadAddr(0)^);
end;

function ucTreeParseUp(FunctionAddress: Pointer): Longint;
begin
   Result := ucSetCallbackE(FunctionAddress, ucTreeUp, ucThreadAddr(0)^);
end;

function ucUpperBound(VarHandle: Longint; Dimension: Longint = 1): Longint;
begin
   Result := ucUpperBoundE(VarHandle, Dimension);
end;

function ucVariableValueStr(VarHandle: Longint): PChar;
begin
   Result := ucGetVariableValueStrE(VarHandle);
end;

function uc_Precedence(var t: ucThreadType): Extended; stdcall;
begin
   Result := ucPrecedence(ucParamStr(1, t), Trunc(ucParam(2, t)), t.ucThread);
end;

function uc_PrecedenceMax(var t: ucThreadType): Extended; stdcall;
begin
   Result := t.PrecedenceMax;
end;

function uc_PrecedenceMin(var t: ucThreadType): Extended; stdcall;
begin
   Result := t.PrecedenceMin;
end;

{$IFNDEF ucAddOnLibrary}
constructor FastMath.Create;
begin
   t := ucNewThread();
end;

destructor FastMath.Destroy;
begin
   ucReleaseThread(t);
end;

procedure FastMath.ucDetachSymbol(Symbol: AnsiString);
begin
   ucDetachSymbolE(Symbol, ucThreadAddr(t)^);
end;

procedure FastMath.ucErrorClear;
begin
   ucErrorClearE(ucThreadAddr(t)^);
end;

procedure FastMath.ucReleaseItem(ItemHandle: Longint);
begin
   ucReleaseItemE(ItemHandle);
end;

procedure FastMath.ucReleaseExprPtr(ExpressionPtr: Pointer);
begin
   ucReleaseExprPtrE(ExpressionPtr);
end;

procedure FastMath.ucReleaseThread(ucThread: Longint);
begin
   if ucThread > 0 then ucReleaseThreadE(ucThread);
end;

procedure FastMath.ucSetThreadIndex(ucThread: Longint);
begin
   if ucThread = 0 then ucReleaseThread(t); // +++ ???
   t := ucThread;
end;

procedure FastMath.ucSetTrigMode(TrigMode: Longint);
begin
   ucSetTrigModeE(TrigMode, ucThreadAddr(t)^);
end;

procedure FastMath.ucSetVariableValue(VarHandle: Longint; vValue: Extended);
begin
   ucSetVariableValueE(VarHandle, vValue);
end;

procedure FastMath.ucSetVariableValueStr(VarHandle: Longint; vValue: AnsiString);
begin
   ucSetVariableValueStrE(VarHandle, vValue);
end;

function FastMath.ucAlias(AssignFrom, AssignTo: AnsiString; aProperties: Longint = 0; Precedence: Longint = 0): Longint;
begin
   Result := ucAliasE(AssignFrom, AssignTo, aProperties, Precedence, ucThreadAddr(t)^);
end;

function FastMath.ucDefine(Definition: AnsiString): Longint;
begin
   Result := ucDefineE(Definition, ucThreadAddr(t)^);
end;

function FastMath.ucDefineConstant(ConstantDef: AnsiString; vProperties: Longint = 0): Longint;
begin
   Result := ucDefineVariableE(ConstantDef, nil, nil, vProperties+ucConstant, t0.ucProperty.DefaultNumericType, ucThreadAddr(t)^);
end;

function FastMath.ucDefineFunction(FunctionDef: AnsiString; FunctionAddress: Pointer = nil; fProperties: Longint = 0; vType: Longint = 0): Longint;
begin
   Result := ucDefineFunctionE(FunctionDef, FunctionAddress, 0, fProperties, vType, ucThreadAddr(t)^);
end;

function FastMath.ucDefineMacro(Definition: AnsiString): Longint;
begin
   Result := ucDefineE('Macro '+Definition, ucThreadAddr(t)^);
end;

function FastMath.ucDefineOperator(OperatorDef: AnsiString; Precedence: Longint = 0; OperatorAddress: Pointer = nil; fProperties: Longint = 0; vType: Longint = 0): Longint;
begin
   Result := ucDefineFunctionE(OperatorDef, OperatorAddress, Precedence, fProperties, vType, ucThreadAddr(t)^);
end;

function FastMath.ucDefineSyntax(Definition: AnsiString): Longint;
begin
   Result := ucDefineE('Syntax '+Definition, ucThreadAddr(t)^);
end;

function FastMath.ucDefineVariable(VariableDef: AnsiString; vProperties: Longint = 0): Longint;
begin
   Result := ucDefineVariableE(VariableDef, nil, nil, vProperties, t0.ucProperty.DefaultNumericType, ucThreadAddr(t)^);
end;

function FastMath.ucAttachVariable(VariableDef: AnsiString; vPtr: Pointer; ExprPtr: Pointer = nil; vProperties: Longint = 0): Longint;
begin
   Result := ucDefineVariableE(VariableDef, vPtr, ExprPtr, vProperties, t0.ucProperty.DefaultNumericType, ucThreadAddr(t)^);
end;

function FastMath.ucErrorHandler(FunctionAddress: Pointer): Longint;
begin
   Result := ucSetCallbackE(FunctionAddress, ucErrorCallback, ucThreadAddr(t)^);
end;

function FastMath.ucEval(Expression: AnsiString): Extended;
begin
   Result := ucEvalE(Expression, ucThreadAddr(t)^);
end;

function FastMath.ucEvalStr(Expression: AnsiString): PChar;
begin
   Result := ucEvalStrE(Expression, ucThreadAddr(t)^);
end;

function FastMath.ucEvaluate(ExpressionPtr: Pointer): Extended;
begin
   Result := ucEvaluateE(ExpressionPtr);
end;

function FastMath.ucEvaluate(ExpressionPtr: Pointer; var ucValue: ucComplexNumberType): Extended;
begin
   Result := ucEvaluateE(ExpressionPtr);
   ucValue.Real := Result;
   ucValue.Imag := ucImagPartE(ucThreadAddr(t)^);
end;

function FastMath.ucEvaluateCplx(ExpressionPtr: Pointer; var ucValue: ucComplexNumberType): Extended;
begin
   Result := ucEvaluateE(ExpressionPtr);
   ucValue.Real := Result;
   ucValue.Imag := ucImagPartE(ucThreadAddr(t)^);
end;

function FastMath.ucEvaluateStr(ExpressionPtr: Pointer): PChar;
begin
   Result := ucEvaluateStrE(ExpressionPtr);
end;

function FastMath.ucGetErrorData(errIndex: Longint): PChar;
begin
   Result := ucGetErrorDataE(errIndex, ucThreadAddr(t)^);
end;

function FastMath.ucError(): Longint;
var tData: ucThreadType;
begin
   ucGetThreadInfo(t, tData);
   Result := tData.ErrorNumber;
end;

function FastMath.ucErrorData(errIndex: Longint): PChar;
begin
   Result := ucGetErrorDataE(errIndex, ucThreadAddr(t)^);
end;

function FastMath.ucGetThreadIndex: Longint;
begin
   Result := t;
end;

function FastMath.ucGetVariableValue(VarHandle: Longint): Extended;
begin
   Result := ucGetVariableValueE(VarHandle);
end;

function FastMath.ucGetVariableValueStr(VarHandle: Longint): PChar;
begin
   Result := ucGetVariableValueStrE(VarHandle);
end;

function FastMath.ucImagPart: Extended;
begin
   Result := ucImagPartE(ucThreadAddr(t)^);
end;

function FastMath.ucNewThread: Longint;
begin
   Result := ucNewThreadE;
end;

function FastMath.ucParse(Expression: AnsiString; LocalVars: AnsiString = ''; ExprType: Longint = 0): Pointer;
begin
   Result := ucParseE(Expression, LocalVars, ExprType, ucThreadAddr(t)^);
end;

function FastMath.ucPrecedence(Op: AnsiString; OpCategory: Longint = 0): Longint;
begin
   Result := ucPrecedenceE(Op, OpCategory, ucThreadAddr(t)^);
end;

function FastMath.ucPreParse(FunctionAddress: Pointer): Longint;
begin
   Result := ucSetCallbackE(FunctionAddress, ucTreePre, ucThreadAddr(t)^);
end;

function FastMath.ucSymbolHandle(ucSymbol: AnsiString; SymbolProperty: Longint = 0): Longint;
begin
   Result := ucSymbolHandleE(ucSymbol, SymbolProperty, ucThreadAddr(t)^);
end;

function FastMath.ucTreeParseDownEarly(FunctionAddress: Pointer): Longint;
begin
   Result := ucSetCallbackE(FunctionAddress, ucTreeDownEarly, ucThreadAddr(t)^);
end;

function FastMath.ucTreeParseDown(FunctionAddress: Pointer): Longint;
begin
   Result := ucSetCallbackE(FunctionAddress, ucTreeDown, ucThreadAddr(t)^);
end;

function FastMath.ucTreeParseUp(FunctionAddress: Pointer): Longint;
begin
   Result := ucSetCallbackE(FunctionAddress, ucTreeUp, ucThreadAddr(t)^);
end;

function FastMath.ucVariableValueStr(VarHandle: Longint): PChar;
begin
   Result := ucGetVariableValueStrE(VarHandle);
end;

function FastMath.ucErrorMessage(ErrorNumber: Longint = -1): PChar;
begin
   Result := ucGetErrorMessage(ErrorNumber, t);
end;

procedure ucReset(var t: ucThreadType);
var
   Between_Mult_Pow, Between_And_Less, Between_Div_Mod: Longint;   
begin
   SetLength(ucItem, 1+t.ucProperty.DefSpace.BaseDefinitionSpace
             +t.ucProperty.DefSpace.DefinitionSpace
             *t.ucProperty.DefSpace.MaxThreadCount);
   
   t.Context := t.Context or ucContext_ucReset;
   t.Handshake.LanguageID := uc_Lang_Delphi;
   t.Handshake.HandleAddr := @ucItem[0];
   ucEvaluateInitE();
   ucResetE(t);

   t.ucProperty.ComplexNumbers := ucTrue;
   t.ucProperty.EscapeChar     := 126; // "~"

   ucDefineConstant('Byte     = uc_Integer8u');
   ucDefineConstant('Integer  = uc_Integer16');
   ucDefineConstant('Word     = uc_Integer16u');
   ucDefineConstant('Long     = uc_Integer32');
   ucDefineConstant('DWord    = uc_Integer32u');
   ucDefineConstant('Quad     = uc_Integer64');
   ucDefineConstant('Currency = uc_Currency');
   ucDefineConstant('CurrencyX= uc_CurrencyExt');
   ucDefineConstant('Single   = uc_FloatSingle');
   ucDefineConstant('Double   = uc_FloatDouble');
   ucDefineConstant('Extended = uc_FloatExtended');
   ucDefineConstant('AsciiZ   = uc_StringLPSTR');
   ucDefineConstant('String   = uc_StringAnsi');
   ucDefineConstant('FixedPt  = Currency');
   ucDefineConstant('Longint  = Long');
   ucDefineConstant('Any      = uc_Any');

   Between_Mult_Pow := (ucPrecedence('*') + ucPrecedence('^')) div 2;
   Between_And_Less := (ucPrecedence('AND') + ucPrecedence('<')) div 2;
   Between_Div_Mod  := (ucPrecedence('/') + ucPrecedence('MOD')) div 2;
   
   t.ucProperty.FunctionProperties := ucFastCallback + ucOverload;

   ucDefineOperator('*[x As Long, y As Long] As Long', ucPrecedence('*'), ucOp(Op_Int_Multiply));
   ucDefineOperator('-[x As Long, y As Long] As Long', ucPrecedence('-'), ucOp(Op_Int_Subtract));
   ucDefineOperator('+[x As Long, y As Long] As Long', ucPrecedence('+'), ucOp(Op_Int_Add));
   ucDefineOperator('/[x As Long, y As Long] As Extended', ucPrecedence('/'), ucOp(Op_Int_DivideToExt));
   ucDefineOperator('\[x As Long, y As Long] As Long', Between_Div_Mod, ucOp(Op_Int_DivideToInt));
   ucDefineOperator('Mod[x As Long, y As Long] As Long', ucPrecedence('MOD'), ucOp(Op_Int_Mod));
   ucDefineOperator('^[x As Extended, y As Long] As Extended', ucPrecedence('^'), ucOp(Op_IntPower));

   ucDefineOperator('*[x As Quad, y As Quad] As Quad', ucPrecedence('*'), ucOp(Op_Qwd_Multiply));
   ucDefineOperator('-[x As Quad, y As Quad] As Quad', ucPrecedence('-'), ucOp(Op_Qwd_Subtract));
   ucDefineOperator('+[x As Quad, y As Quad] As Quad', ucPrecedence('+'), ucOp(Op_Qwd_Add));
   ucDefineOperator('/[x As Quad, y As Quad] As Extended', ucPrecedence('/'), ucOp(Op_Qwd_DivideToExt));
   ucDefineOperator('\[x As Quad, y As Quad] As Quad', Between_Div_Mod, ucOp(Op_Qwd_DivideToInt));

   ucDefineOperator('*[x As FixedPt, y As FixedPt] As FixedPt', ucPrecedence('*'), ucOp(Op_Cur_Multiply));
   ucDefineOperator('-[x As FixedPt, y As FixedPt] As FixedPt', ucPrecedence('-'), ucOp(Op_Cur_Subtract));
   ucDefineOperator('+[x As FixedPt, y As FixedPt] As FixedPt', ucPrecedence('+'), ucOp(Op_Cur_Add));
   ucDefineOperator('/[x As FixedPt, y As FixedPt] As Extended', ucPrecedence('/'), ucOp(Op_Cur_DivideToExt));
   
   ucDefineFunction('Fact(x As Long)', ucOp(Op_Func_Factorial));
   ucDefineFunction('GetFPU(x As Long) As Long', ucOp(Op_Func_GetFPU));
   ucDefineFunction('SetFPU(x As Long)', ucOp(Op_Func_SetFPU));
   ucDefineFunction('SetOrFPU(x As Long)', ucOp(Op_Func_SetOrFPU));
   ucDefineFunction('SetXorFPU(x As Long)', ucOp(Op_Func_SetXorFPU));
   ucDefineFunction('RestoreFPU(x As Long)', ucOp(Op_Func_RestoreFPU));
   ucDefineFunction('Ceil(x)', ucOp(Op_Func_Ceil));
   ucDefineFunction('Not(x)', ucOp(Op_Func_Not));

   // String relational (comparison) operators
   t.ucProperty.DefaultPrecedence := ucPrecedence('<');
   t.ucProperty.FunctionProperties := ucDynamic + ucLocked + ucOverride + ucOverload;
   ucDefineFunction('StrCmp(a As String, b As String) = StrCmp(0, 0, a, b)');
   ucDefineOperator(' =[a As String, b As String] = StrCmp(a, b) = 2');
   ucDefineOperator('<>[a As String, b As String] = StrCmp(a, b) <> 2');
   ucDefineOperator(' >[a As String, b As String] = StrCmp(a, b) = 3');
   ucDefineOperator('>=[a As String, b As String] = StrCmp(a, b) > 1');
   ucDefineOperator(' <[a As String, b As String] = StrCmp(a, b) = 1');
   ucDefineOperator('<=[a As String, b As String] = StrCmp(a, b) < 3');

   t.ucProperty.FunctionProperties := ucInline+ucLocked+ucOverride+ucOverload;

   ucDefineFunction('Precedence(Operator$, OpCategory = 0)', @uc_Precedence);
   ucDefineFunction('PrecedenceMax()', @uc_PrecedenceMax);
   ucDefineFunction('PrecedenceMin()', @uc_PrecedenceMin);

   ucDefineMacro('.[x]  ::= (0.x) ::: Precedence = PrecedenceMax');
   ucDefineMacro(' [x]. ::= (x.0) ::: Precedence = PrecedenceMax');

   // Hex, binary, octal notations.  Includes support for non-integers
   ucDefineSyntax('#b (Number:"[0-1]+") [(Frac:"\.[0-1]+"=)] ::= BaseConvert(~qNumber~jFrac~q, 2) ::: Precedence = PrecedenceMax');
   ucDefineSyntax('#o (Number:"[0-7]+") [(Frac:"\.[0-7]+"=)] ::= BaseConvert(~qNumber~jFrac~q, 8) ::: Precedence = PrecedenceMax');
   ucDefineSyntax('#h (Number:"[0-9A-F]+") [(Frac:"\.[0-9A-F]+"=)] ::= BaseConvert(~qNumber~jFrac~q, 16) ::: Precedence = PrecedenceMax');
   ucDefineSyntax('#i ::= ~i');

   // Alternative notations
//   ucDefineOperator('  &[a,b] = a and b', ucPrecedence('and'));
//   ucDefineOperator('  |[a,b] = a or  b', ucPrecedence('or '));
//   ucDefineOperator(' =<[a,b] = a <=  b', ucPrecedence('<= '));
//   ucDefineOperator(' =>[a,b] = a >=  b', ucPrecedence('>= '));
   ucDefineOperator(' ==[a,b] = a  =  b', ucPrecedence('='));
   ucDefineOperator('DIV[a,b] = a  \  b', ucPrecedence('\'));
   ucDefineOperator('**[a,b]  = a  ^  b', ucPrecedence('^'));
   ucDefineMacro('&[TextA,TextB] ::= TextA + TextB');
   ucDefineOperator('==[TextA$,TextB$] = TextA = TextB', ucPrecedence('='));
   ucDefineOperator('Imp[a, b] = (Not(a)) Or b', ucPrecedence('OR')-1);
   ucDefineOperator('Eqv[a, b] = (a and b) Or (not(a) and not(b))', ucPrecedence('OR')-1);
   ucDefineFunction('Concatenate$(a$,b$) = a + b');
   ucDefineFunction('Sign(x) = Sgn(x)');
   ucDefineFunction('Sqrt(x) = Sqr(x)');
   ucDefineFunction('Ln(x)   = Log(x)');
   ucDefineFunction('Atan(x) = Atn(x)');
   ucDefineOperator('[x]! = fact(x)', t.PrecedenceMax);
   ucDefineOperator('[x]% = x/100', t.PrecedenceMax);
   ucDefineOperator('%[x,y] = x mod y', ucPrecedence('MOD'));
   ucDefineOperator('-[x] = 0-x', Between_Mult_Pow);
   ucDefineOperator('+[x] = 0+x', Between_Mult_Pow);
   ucDefineOperator('Not[x] = Not(x)', Between_And_Less);

   // FPU routines
   ucDefineFunction('GetFPU() = GetFPU(0)');
   ucDefineFunction('RestoreFPU() = RestoreFPU(0)');
   ucAlias('GetFPU', 'GetFPU(0)');
   ucAlias('RestoreFPU', 'RestoreFPU(0)');
   ucAlias('SetFPU(setting, Expression)',   '(SetFPU(setting)+(Expression)+RestoreFPU(0))');
   ucAlias('SetOrFPU(setting, Expression)', '(SetOrFPU(setting)+(Expression)+RestoreFPU(0))');
   ucAlias('SetXorFPU(setting, Expression)','(SetXorFPU(setting)+(Expression)+RestoreFPU(0))');
   // +++ RestoreFPU(0) instead of RestoreFPU in alias.  Must fix.

   ucDefineFunction('Inf() = SetOrFPU(4, 1/0)', nil, ucDynamic);
   ucDefineFunction('NaN() = SetOrFPU(1, 0/0)', nil, ucDynamic);

   // Hyperbolic functions
   ucDefineFunction('Sinh(x)  = (exp(x) - exp(-x)) / 2');
   ucDefineFunction('Cosh(x)  = (exp(x) + exp(-x)) / 2');
   ucDefineFunction('Tanh(x)  = (exp(2x) - 1) / (exp(2x) + 1)');
   ucDefineFunction('Coth(x)  = (exp(2x) + 1) / (exp(2x) - 1)');
   ucDefineFunction('Sech(x)  = 2 / (exp(x) + exp(-x))');
   ucDefineFunction('Csch(x)  = 2 / (exp(x) - exp(-x))');
   ucDefineFunction('Asinh(x) = Log(x + Sqr(x ^ 2 + 1))');
   ucDefineFunction('Acosh(x) = Log(x + Sqr(x ^ 2 - 1))');
   ucDefineFunction('Atanh(x) = Log((1 + x) / (1 - x)) / 2');
   ucDefineFunction('Acoth(x) = Log((x + 1) / (x - 1)) / 2');
   ucDefineFunction('Asech(x) = Log((1 + Sqr(1 - x ^ 2)) / x)');
   ucDefineFunction('Acsch(x) = Log(1 / x + Sqr(1 + x ^ 2) / Abs(x))');

   // A few constants needed in some definitions further down
   ucDefineConstant('Pi     = Atn(1) * 4');
   ucDefineConstant('HalfPI = Pi/2');

   // Other functions
   ucDefineFunction('Sec(x)   = 1 / Cos(x)');
   ucDefineFunction('Csc(x)   = 1 / Sin(x)');
   ucDefineFunction('Cot(x)   = 1 / Tan(x)');
   ucDefineFunction('Asin(x)  = SetOrFPU(4, Atn(x / Sqr(1 - x * x)))'); // +++ make provisions for complex
   ucDefineFunction('Acos(x)  = HalfPI - Asin(x)');
   ucDefineFunction('Log2(x)  = Log(x) / Log(2)');
   ucDefineFunction('Exp2(x)  = 2 ^ x');
   ucDefineFunction('Log10(x) = Log(x) / Log(10)');
   ucDefineFunction('Exp10(x) = 10 ^ x');

   // Complex operators & functions
   t.ucProperty.FunctionProperties := ucDynamic+ucLocked+ucOverride+ucOverload+ucComplex;
   ucDefineOperator('+[x] = x.Real + x.Imag*#i', Between_Mult_Pow);
   ucDefineOperator('-[x] = (-x.Real) + (-x.Imag)*#i', Between_Mult_Pow);

   ucDefineOperator('+[x,y] = x.Real+y.Real + (x.Imag+y.Imag)*#i', ucPrecedence('+'));
   ucDefineOperator('*[x,y] = (x.Real*y.Real - x.Imag*y.Imag)+(x.Real*y.Imag + x.Imag*y.Real)*#i', ucPrecedence('*'));
   ucDefineOperator('-[x,y] = (x.Real-y.Real) + (x.Imag-y.Imag)*#i', ucPrecedence('-'));
   ucDefineOperator('/[x,y] = x.Real*y.Real/(y.Real^2+y.Imag^2) + x.Imag*y.Imag'
   + '/(y.Real^2+y.Imag^2) + (x.Imag*y.Real/(y.Real^2+y.Imag^2)-x.Real*y.Imag'
   + '/(y.Real^2+y.Imag^2))*#i', ucPrecedence('/'));

   ucDefineFunction('Abs(x) = Sqr(x.Real^2 + x.Imag^2) + 0*#i');
   ucDefineFunction('Sgn(x) = x.Real/Abs(x) + x.Imag/Abs(x)*#i');
   ucDefineFunction('Sin(x) = Sin(x.Real)*Cosh(x.Imag) + Cos(x.Real)*Sinh(x.Imag)*#i');
   ucDefineFunction('Cos(x) = Cos(x.Real)*Cosh(x.Imag) - Sin(x.Real)*Sinh(x.Imag)*#i');
   ucDefineFunction('Phase(x) = iif(x.Imag=0, iif(x.Real>0, 0, pi), atn(x.Imag/x.Real)) + 0*#i');
   ucDefineFunction('Conj(x)  = x.Real - x.Imag*#i');
   ucDefineFunction('Exp(x)   = Exp(x.Real)*Cos(x.Imag) + Exp(x.Real)*Sin(x.Imag)*#i');
   ucDefineFunction('Imag(x)  = x.Imag + 0*#i');
   ucDefineFunction('Real(x)  = x.Real + 0*#i');

   ucDefineFunction('Sqr(x) = Sqr(Sqr(x.Real^2 + x.Imag^2)+x.Real)/Sqr(2)'
   + '+(Sqr(Sqr(x.Real^2+x.Imag^2)-x.Real)*iif(x.Imag < 0, -1, 1)/Sqr(2))*#i');
   ucDefineFunction('Sqrt(x)= Sqr(x.Real+x.Imag*#i)');

   ucDefineFunction('Log(x) = iif(x.Real=0 and x.Imag=0, -Inf,'
   + 'Log(x.Real^2 + x.Imag^2)/2)'
   + '+iif(x.Real=0 and x.Imag=0, 0, iif(x.Real>0 and x.Imag=0, 0,'
   + 'iif(x.Imag=0, pi, pi*sgn(x.Imag)/2-Atn(x.Real/x.Imag))))*#i');

   ucDefineFunction('_ZeroToZero() = 1', nil, ucDynamic+ucOverwrite);
// ucDefineFunction('_ZeroToZero() = 0', nil, ucDynamic+ucOverwrite);
// ucDefineFunction('_ZeroToZero() = NaN', nil, ucDynamic+ucOverwrite);
// ucDefineFunction('_ZeroToZero() = SetError(1, "0^0 Error")', nil, ucDynamic+ucOverwrite);

   ucDefineOperator('^[x,y] = iif(x.Imag = 0, iif(x.Real > 0,'
   + 'x.Real^y.Real*Cos(y.Imag*Log(x.Real)),'
   + 'iif(x.Real=0 and y.Real=0 and y.Imag=0, _ZeroToZero(),'
   + '(-x.Real)^y.Real*COS(y.Imag*LOG(-x.Real)+pi*y.Real)/EXP(pi*y.Imag))),'
   + 'EXP(y.Imag*ATN(x.Real/x.Imag))*SQR((x.Imag^2+x.Real^2)^y.Real)'
   + '*COS(pi*y.Real*sgn(x.Imag)/2-y.Real*ATN(x.Real/x.Imag)+y.Imag'
   + '*LOG(x.Imag^2+x.Real^2)/2)/SQR(EXP(pi*y.Imag*Sgn(x.Imag))))-55+55'
   + '+iif(x.Imag = 0, iif(x.Real > 0,'
   + 'x.Real^y.Real*SIN(y.Imag*LOG(x.Real)),'
   + 'iif(x.Real=0 and y.Real=0 and y.Imag=0, 0,'
   + '(-x.Real)^y.Real*SIN(y.Imag*LOG(-x.Real)+pi*y.Real)/EXP(pi*y.Imag))),'
   + 'EXP(y.Imag*ATN(x.Real/x.Imag))*SQR((x.Imag^2+x.Real^2)^y.Real)'
   + '*SIN(pi*y.Real*sgn(x.Imag)/2-y.Real*ATN(x.Real/x.Imag)+y.Imag'
   + '*LOG(x.Imag^2+x.Real^2)/2)/SQR(EXP(pi*y.Imag*Sgn(x.Imag)))-55+55)*#i',
     ucPrecedence('^'));

   t.ucProperty.DefaultPrecedence := Between_Div_Mod; // arbitrarily chosen
   t.Context := t.Context xor ucContext_ucReset;
   t.ucProperty.FunctionProperties :=
       ucDynamic
     + ucOverload
     + ucOverwrite
     + ucComplex
     + ucRecursion;
end;
{$ENDIF}

{$IFNDEF ucAltErrMsg}
function ucGetErrorMessage(ErrorNumber: Longint = -1; t: Longint = 0): PChar;
var
   ErrMsg: AnsiString;
begin
   if ErrorNumber = -1 then ErrorNumber := t0.ErrorNumber;

   case ErrorNumber of
      ucErr_None:                ErrMsg := ''; // No Error
      ucErr_ucErrorData:         ErrMsg := ucGetErrorData(0, t);
      ucErr_SyntaxError:         ErrMsg := 'Syntax error';
      ucErr_MismatchedParenth:   ErrMsg := 'Mismatched parenthesis';
      ucErr_MismatchedQuotes:    ErrMsg := 'Mismatched quotes';
      ucErr_UndefinedVariable:   ErrMsg := 'Undefined variable';
      ucErr_UndefinedFunction:   ErrMsg := 'Undefined function';
      ucErr_IncorrectArgType:    ErrMsg := 'Incorrect argument type';
      ucErr_InvalidNumberOfArgs: ErrMsg := 'Invalid number of function arguments';
      ucErr_InvalidOperandType:  ErrMsg := 'Invalid operand type';
      ucErr_InvalidVariableName: ErrMsg := 'Invalid variable name';
      ucErr_ImplicitMult:        ErrMsg := 'Implicit multiplication error';
      ucErr_IncompleteBlock:     ErrMsg := 'Incomplete syntax block';

      ucErr_InvalidOperation:    ErrMsg := 'Invalid floating point operation';
      ucErr_DenormalOperand:     ErrMsg := 'Denormal floating point operand';
      ucErr_DivisionByZero:      ErrMsg := 'Division by 0';
      ucErr_Overflow:            ErrMsg := 'Floating point overflow';
      ucErr_Underflow:           ErrMsg := 'Floating point underflow';
      ucErr_InexactResult:       ErrMsg := 'Inexact floating point result';
      ucErr_SubscriptOutOfRange: ErrMsg := 'Array subscript out of range';
      ucErr_StackOverflow:       ErrMsg := 'Stack overflow';

      ucErr_OperationDenied:     ErrMsg := 'Cannot perform requested operation';
      ucErr_PropertyConflict:    ErrMsg := 'Definition property conflict';
      ucErr_ConstantWriteDenied: ErrMsg := 'Assignment to a constant not allowed';
      ucErr_FunctionDefError:    ErrMsg := 'Invalid function or operator definition';
      ucErr_FunctionLocked:      ErrMsg := 'Function or operator is locked';

      ucErr_DefinitionSpaceFull: ErrMsg := 'Definition space is full';
      ucErr_ThreadInitFailed:    ErrMsg := 'Thread initialization failed';
      ucErr_UnsupportedFeature:  ErrMsg := 'Requires advanced edition of UCalc FMP';
      // numbers up to 500 are reserved for future use by UCalc
      // Use Case 501 and beyond for user-defined errors
   else
      ErrMsg:= 'Unknown error';
   end;

   If ErrorNumber > 1 Then ErrMsg := 'Error #' + FloatToStr(ErrorNumber) + ': '  + ErrMsg;

   ucGetErrorMessage := PChar(ErrMsg);
end;
{$ENDIF}

function ucErrorMessage(ErrorNumber: Longint = -1; t: Longint = 0): PChar;
begin
   Result := ucGetErrorMessage(ErrorNumber, t);
end;

{$IFNDEF ucAddOnLibrary}
var HostFPUWord: Word;
initialization
   HostFPUWord := Default8087CW;
   ucGetThreadInfo(0, t0);
   ucSetThreadAddr(0, t0);
   ucReset(t0);
   t0.ucProperty.FPU_ControlWord_HostProgram := HostFPUWord;
end.
{$ENDIF}
