' -------------------------------------------------------------------------
' ucalcpb.bas : PowerBASIC module file
' Designed to work with PB/DLL 6.1 or above, and PB/CC 2.1 or above
'
' UCalc Fast Math Parser (Beta N)
' Copyright (c) 1998-2005 by Daniel Corbier.  All rights reserved.
' http://www.ucalc.com/mathparser
' -------------------------------------------------------------------------

'Option Explicit '+++

Type ucDefinitionSpaceType
   BaseDefinitionSpace  As Long
   DefinitionSpace      As Long
   InlineBuffer         As Long
   MaxErrorData         As Long
   MaxThreadCount       As Long
End Type

Type ucForInternalUseType
   LanguageID           As Long
   ucOperator           As Dword
   ucEvalAddr           As Dword
   ucEvaluateAddr       As Dword
   ucThreadAddr         As Dword
   ParamListAddr        As Dword
   HandleAddr           As Dword
   ReleaseExprPtr       As Dword
   TreeIndex            As Long
   TreeParamInfoAddress As Dword
   HaltEvaluation       As Long
   ThreadStatus         As Long
   PrevOperationCount   As Long
   PrevItemCount        As Long
   CompareOld           As Long
   IsStringValue        As Long
   ucStack              As Long
   Infinity             As Long
   TrigFactor           As Ext Ptr
   TrigFactorInv        As Ext Ptr

   DataGroup1           As Dword
   DataGroup1b          As Dword
   DataGroup1c          As Dword
   DataGroup2           As Dword
   DataGroup2b          As Dword
   DataGroup2c          As Dword
   DataGroup3           As Dword
   DataGroup3b          As Dword
   DataGroup3c          As Dword
   DataGroup4           As Dword
   DataGroup4b          As Dword
   DataGroup4c          As Dword
   StringPtr1           As Dword
End Type

Type ucPropertyType
   DefaultNumericType          As Long
   DefaultExternalStringType   As Long
   FPU_ControlWord_UCalc       As Long
   FPU_ControlWord_HostProgram As Long
   FPU_Turbo                   As Long
   CaseSensitive               As Long
   ComplexNumbers              As Long
   ImagSymbol                  As Long
   RelationalTrue              As Long
   ImplicitMultiply            As Long
   MaxParamPerFunction         As Long
   ArrayProperties             As Long
   FunctionProperties          As Long
   StackSize                   As Long
   DefaultPrecedence           As Long
   OutputDigits                As Long
   QuoteChar                   As Long
   QuoteCharSynonym            As Long
   ArgumentSeparator           As Long
   DecimalSeparator            As Long
   EscapeChar                  As Long
   DefSpace                    As ucDefinitionSpaceType
End Type

Union DataByTypeType
   ValuePtr    As Dword

   ValueByte   As Byte      Ptr
   ValueInt    As Integer   Ptr
   ValueWrd    As Word      Ptr
   ValueLng    As Long      Ptr
   ValueDwd    As Dword     Ptr
   ValueQuad   As Quad      Ptr

   ValueSng    As Single    Ptr
   ValueDbl    As Double    Ptr
   ValueExt    As Extended  Ptr

   ValueCur    As Currency  Ptr
   ValueCux    As CurrencyX Ptr

   ValueStr    As String    Ptr
   ValueAsciiz As Asciiz    Ptr
End Union

Type ucHandleType
   ucThread          As Long
   Info              As Dword
   DataType          As Long
   DataByTypeType
   Cplx              As DataByTypeType
   SymbolName        As String Ptr
   SortedHandle      As Long
   OverloadHandle    As Long
   OpPrecedence      As Long
   ReqParamCount     As Integer
   ParamCount        As Integer
   CallbackPtr       As Dword
   FuncArgs          As String Ptr
   FuncArgHandleData As String Ptr
   FuncArgTypeList   As String Ptr
   VarValueStr       As String Ptr

   DataGroup1        As Dword
   DataGroup1b       As Dword
   DataGroup1c       As Dword
   DataGroup2        As Dword
   DataGroup2b       As Dword
   DataGroup2c       As Dword
   StringPtr1        As Dword
   StringPtr2        As Dword
   StringPtr3        As Dword
   StringPtr4        As Dword
   StringPtr5        As Dword

   UserData1         As Dword
   UserData2         As Dword
   UserData3         As Dword
   UserData4         As Dword
End Type

Type ucThreadType
   ucThread          As Long
   Context           As Long
   ParamCount        As Long
   ErrorNumber       As Long
   TreeDirection     As Long
   DefSpace          As Long
   OperationCount    As Long
   ItemCount         As Long
   Sequence          As Long
   SequenceLast      As Long
   ExprHasChanged    As Long
   IsFunctionParse   As Long
   ComplexStatus     As Long
   ucParamAddr       As Dword
   ucParamHandleAddr As Dword
   SymbolStart       As Long
   SymbolEnd         As Long
   PrecedenceMax     As Long
   PrecedenceMin     As Long
   VarHandle         As Long
   OpHandle          As Long
   StringWasReturned As Long
   ReturnString      As String Ptr
   ReturnValue       As Ext Ptr
   ReturnComplex     As Ext Ptr
   ucProperty        As ucPropertyType
   Handshake         As ucForInternalUseType
   User1             As Long
   User2             As Long
   User3             As Long
   User4             As Long
End Type

Type ucComplexNumberType
   Real As Ext
   Imag As Ext
End Type

%ucTrue  = 1
%ucFalse = 0

' LanguageID
%uc_Lang_Unsupported    = 0
%uc_Lang_Debug          = 1
%uc_Lang_PowerBasic     = 2
%uc_Lang_Delphi         = 3
%uc_Lang_VisualBasic    = 4
%uc_Lang_VisualBasicNET = 5
%uc_Lang_CPlusPlus      = 6

' Type constants for vType arg of ucDefineFunction and ucDefineVariable
%ucType_Integer8      = 1
%ucType_Integer8u     = 2
%ucType_Integer16     = 3
%ucType_Integer16u    = 4
%ucType_Integer32     = 5
%ucType_Integer32u    = 6
%ucType_Integer64     = 7
%ucType_Integer64u    = 8
%ucType_Currency      = 9
%ucType_CurrencyExt   = 10
%ucType_FloatSingle   = 11
%ucType_FloatDouble   = 12
%ucType_FloatExtended = 13
%ucType_StringBSTR    = 14  ' For VB
%ucType_StringLPSTR   = 15  ' (ASCIIZ) C++ char default
%ucType_StringAnsi    = 16  ' PowerBASIC and Delphi default
%ucType_Any           = 17

%ucType_Long          = %ucType_Integer32
%ucType_Double        = %ucType_FloatDouble
%ucType_Extended      = %ucType_FloatExtended
%ucType_FixedPt       = %ucType_Currency
%ucType_Quad          = %ucType_Integer64
%ucType_AnsiString    = %ucType_StringAnsi
%ucType_AsciiZ        = %ucType_StringLPSTR

' Context options (t.Context)
%ucContext_Escape            = 1
%ucContext_Try               = 2
%ucContext_ucDefine          = 4
%ucContext_ucDefineFunction  = 8
%ucContext_ucDefineVariable  = 16
%ucContext_ucEval            = 32
%ucContext_ucEvalStr         = 64
%ucContext_ucEvaluateStr     = 128
%ucContext_ucParse           = 256
%ucContext_ucParseOrucEval   = 512
%ucContext_ucReset           = 1024
%ucContext_Reserved          = 2048

' Trig mode options for ucSetTrigMode
%ucRadians = 1
%ucDegrees = 2
%ucGrads   = 3
%ucCycles  = 4

' Options for ucArrayInfo,
' and the first two for ucProperty.ArrayProperties
%ucArray_BoundsCheck    = 1
%ucArray_ColumnMajor    = 2
%ucArray_DataAddress    = 3
%ucArray_DataType       = 4
%ucArray_DimensionCount = 5
%ucArray_ElementCount   = 6
%ucArray_ElementSize    = 7

' FPU control word values for ucProperty.FPU_ControlWord,
' and SetFPU(), SetOrFPU, SetXorFPU, GetFPU(), RestoreFPU
%ucFPU_ErrMask_InvalidOp     = 1
%ucFPU_ErrMask_DenormalOp    = 2
%ucFPU_ErrMask_ZeroDivide    = 4
%ucFPU_ErrMask_Overflow      = 8
%ucFPU_ErrMask_Underflow     = 16
%ucFPU_ErrMask_PrecisionLoss = 32

%ucFPU_Precision_Single      = 0
%ucFPU_Precision_Double      = 512
%ucFPU_Precision_Extended    = 768

%ucFPU_Round_Even            = 0
%ucFPU_Round_Down            = 1024
%ucFPU_Round_Up              = 2048
%ucFPU_Round_Truncate        = 3072

' For ucProperty.FPU_Turbo
%ucFPU_Safe      = 0
%ucFPU_SkipSet   = 1
%ucFPU_SkipReset = 65536

' Options for fProperties parameter of ucDefineFunction
' and for and for ucProperty.FunctionProperties
%ucLeftToRight  = 0
%ucInline       = 1
%ucDynamic      = 2
%ucFastCallback = 4
%ucLocal        = 8
%ucComplex      = 16
%ucLocked       = 32
%ucConstant     = %ucLocked
%ucOverride     = 64
%ucOverwrite    = 128
%ucOverload     = 256
%ucRecursion    = 512
%ucBootstrap    = 1024
%ucDelaySyntax  = 1024
%ucRightToLeft  = 2048

' For ucTreeProperty, ucTreeNodeProperty,
' ucTreeParamProperty, and ucTreeParentProperty
%ucIsInline          = 0
%ucIsDynamic         = 1
%ucIsFastCallback    = 2
%ucIsLocal           = 3
%ucIsComplex         = 4
%ucIsLocked          = 5
%ucIsConstant        = %ucIsLocked
%ucIsOverride        = 6
%ucIsOverwrite       = 7
%ucIsOverload        = 8
%ucIsRecursion       = 9
%ucIsBootstrap       = 10
%ucIsDelaySyntax     = 10
%ucIsRightToLeft     = 11

%ucIsArray           = 12

%ucIsFunction        = 13
%ucIsOperator        = 14
%ucIsInfixOp         = 15
%ucIsPostfixOp       = 16
%ucIsPrefixOp        = 17
%ucIsVariable        = 18
%ucIsJITVariable     = 19
%ucIsJITExpression   = 20
%ucIsString          = 21
%ucIsExternal        = 22
%ucIsActive          = 23
%ucIsItemCount       = 24
%ucIsSorted          = 25
%ucIsMaybeComplex    = 26
%ucIsComplexAltDef   = 27
%ucIsArgTypeOverload = 28
%ucIsSyntax          = 29
%Reserved1           = 30
%Reserved2           = 31

%ucInfixOp   = %ucIsInfixOp
%ucPostfixOp = %ucIsPostfixOp
%ucPrefixOp  = %ucIsPrefixOp

' For ucElementCategory
%ucIsOperand    = 32
'%ucIsFunction  = (already defined)
'%ucIsInfixOp   = (already defined)
'%ucIsPrefixOp  = (already defined)
'%ucIsPostfixOp = (already defined)

%ucDataType    = 40
%ucResultPtr   = 41
%ucSignPos     = 42
%ucElementCat  = 43
%ucTreeIndex   = 44
%ucErrorNumber = 45
%ucOpHandle    = 46
%ucFuncHandle  = %ucOpHandle
%ucExprPtr     = 47
%ucCodeAddress = 48
%ucCurrentPtr  = 49
%ucParentPtr   = 50
%ucDataHandle  = 51
%ucMiscData    = 52
%ucUserData1   = 53
%ucUserData2   = 54
%ucUserData3   = 55
%ucUserData4   = 56

%ucTreePre  = 5
%ucTreeDownEarly = 6
%ucTreeDown = 7
%ucTreeUp   = 8

%ucErrorCallback = 9

%ucNoChange    = 0
%ucMinorChange = 1
%ucMajorChange = 2

' For ucSetErrorData / ucGetErrorData
%ucNewIndex = -2
%ucNewUniqueIndex  = -1
%ucAllErrorData    = -1

' Error values
%ucErr_None                = 0
%ucErr_ucErrorData         = 1
%ucErr_SyntaxError         = 2
%ucErr_MismatchedParenth   = 3
%ucErr_MismatchedQuotes    = 4
%ucErr_UndefinedVariable   = 5
%ucErr_UndefinedFunction   = 6
%ucErr_IncorrectArgType    = 7
%ucErr_InvalidNumberOfArgs = 8
%ucErr_InvalidOperandType  = 9
%ucErr_InvalidVariableName = 10
%ucErr_ImplicitMult        = 11
%ucErr_IncompleteBlock     = 12

%ucErr_InvalidOperation    = 20
%ucErr_DenormalOperand     = 21
%ucErr_DivisionByZero      = 22
%ucErr_Overflow            = 23
%ucErr_Underflow           = 24
%ucErr_InexactResult       = 25
%ucErr_SubscriptOutOfRange = 35
%ucErr_StackOverflow       = 36

%ucErr_OperationDenied     = 40
%ucErr_PropertyConflict    = 41
%ucErr_ConstantWriteDenied = 42
%ucErr_FunctionDefError    = 43
%ucErr_FunctionLocked      = 44

%ucErr_DefinitionSpaceFull = 60
%ucErr_ThreadInitFailed    = 61
%ucErr_UnsupportedFeature  = 62

%Op_Func_GetFPU     = 36
%Op_Func_SetFPU     = 37
%Op_Func_SetOrFPU   = 38
%Op_Func_SetXorFPU  = 39
%Op_Func_RestoreFPU = 40
%Op_Func_Ceil       = 41
%Op_Func_Factorial  = 42
%Op_Func_Not        = 43
%Op_Cur_Add         = 44
%Op_Cur_Multiply    = 45
%Op_Cur_Subtract    = 46
%Op_Cur_DivideToExt = 47
%Op_Qwd_Add         = 48
%Op_Qwd_Multiply    = 49
%Op_Qwd_Subtract    = 50
%Op_Qwd_DivideToInt = 51
%Op_Qwd_DivideToExt = 52
%Op_Int_Add         = 53
%Op_Int_Multiply    = 54
%Op_Int_Subtract    = 55
%Op_Int_DivideToInt = 56
%Op_Int_DivideToExt = 57
%Op_Int_Mod         = 58
%Op_IntPower        = 59

Global t0 As ucThreadType
Global ucInitialized As Long
Global ucItem() As ucHandleType

Declare Sub ucDetachSymbolE Lib "ucalcfmp.dll" Alias "ucDetachSymbolE" (Symbol As Asciiz, t As ucThreadType)
Declare Sub ucErrorClearE Lib "ucalcfmp.dll" Alias "ucErrorClearE" (ByVal t As Dword)
Declare Sub ucLicenseE Lib "ucalcfmp.dll" Alias "ucLicenseE" (SerialNumber As Asciiz)
Declare Sub ucReleaseItemE Lib "ucalcfmp.dll" Alias "ucReleaseItemE" (ByVal ItemHandle As Long)
Declare Sub ucReleaseThreadE Lib "ucalcfmp.dll" Alias "ucReleaseThreadE" (ByVal ucThread As Long)
Declare Sub ucResetE Lib "ucalcfmp.dll" Alias "ucResetE" (t As ucThreadType)
Declare Sub ucReturnE Lib "ucalcfmp.dll" Alias "ucReturnE" (ByVal vValue As Ext, t As ucThreadType)
Declare Sub ucReturnCplxE Lib "ucalcfmp.dll" Alias "ucReturnCplxE" (ByVal vValue As Ext, t As ucThreadType)
Declare Sub ucReturnStrE Lib "ucalcfmp.dll" Alias "ucReturnStrE" (Expression As Asciiz, t As ucThreadType)
Declare Sub ucSetErrorDataE Lib "ucalcfmp.dll" Alias "ucSetErrorDataE" (ByVal errIndex As Long, ByVal vValue As String, t As ucThreadType)
Declare Sub ucSetParamStrE Lib "ucalcfmp.dll" Alias "ucSetParamStrE" (ByVal ParamIndex As Long, Argument As Asciiz, t As ucThreadType)
Declare Sub ucSetThreadAddrE Lib "ucalcfmp.dll" Alias "ucSetThreadAddrE" (ByVal ucThread As Long, t As ucThreadType)
Declare Sub ucSetTrigModeE Alias "ucSetTrigModeE" (ByVal TrigMode As Long, t As ucThreadType)
Declare Sub ucSetVariableValueE Lib "ucalcfmp.dll" Alias "ucSetVariableValueE" (ByVal VarHandle As Long, ByVal vValue As Ext)
Declare Sub ucSetVariableValueStrE Lib "ucalcfmp.dll" Alias "ucSetVariableValueStrE" (ByVal VarHandle As Long, ByVal vValue As String)
Declare Sub ucTreeSetParamE Lib "ucalcfmp.dll" Alias "ucTreeSetParamE" (ByVal pIndex As Long, pData As Asciiz, t As ucThreadType)
Declare Function ucAliasE Lib "ucalcfmp.dll" Alias "ucAliasE" (AssignFrom As Asciiz, AssignTo As Asciiz, ByVal aProperties As Long, ByVal Precedence As Long, ByVal t As Dword) As Long
Declare Function ucArrayInfoE Lib "ucalcfmp.dll" Alias "ucArrayInfoE" (ByVal VarHandle As Long, ByVal InfoOption As Long) As Dword
Declare Function ucDefineE Lib "ucalcfmp.dll" Alias "ucDefineE" (Definition As Asciiz, ByVal t As Dword) As Long
Declare Function ucDefineFunctionE Lib "ucalcfmp.dll" Alias "ucDefineFunctionE" (FunctionDef As Asciiz, ByVal FunctionAddress As Dword, ByVal Precedence As Long, ByVal fProperties As Long, ByVal vType As Long, ByVal t As Dword) As Long
Declare Function ucDefineVariableE Lib "ucalcfmp.dll" Alias "ucDefineVariableE" (VariableDef As Asciiz, ByVal vPtr As Dword, ByVal Expr As Dword, ByVal vProperties As Long, ByVal vType As Long, ByVal t As Dword) As Long
Declare Function ucEvalE Lib "ucalcfmp.dll" Alias "ucEvalE" (Expression As Asciiz, ByVal t As Dword) As Ext
Declare Function ucEvalStrE Lib "ucalcfmp.dll" Alias "ucEvalStrE" (Expression As Asciiz, ByVal t As Dword) As String
Declare Function ucEvaluateStrE Lib "ucalcfmp.dll" Alias "ucEvaluateStrE" (ByVal ExpressionPtr As Long) As String
Declare Function ucGetErrorDataE Lib "ucalcfmp.dll" Alias "ucGetErrorDataE" (ByVal errIndex As Long, ByVal t As Dword) As String
Declare Function ucGetParamE Lib "ucalcfmp.dll" Alias "ucGetParamE" (ByVal pIndex As Long, t As ucThreadType) As Ext
Declare Function ucGetParamCplxE Lib "ucalcfmp.dll" Alias "ucGetParamCplxE" (ByVal pIndex As Long, t As ucThreadType) As Ext
Declare Function ucGetParamStrE Lib "ucalcfmp.dll" Alias "ucGetParamStrE" (ByVal ParamIndex As Long, t As ucThreadType) As String
Declare Function ucGetThreadInfoE Lib "ucalcfmp.dll" Alias "ucGetThreadInfoE" (ByVal ucThread As Long, t As ucThreadType) As Dword
Declare Function ucGetVariableValueE Lib "ucalcfmp.dll" Alias "ucGetVariableValueE" (ByVal VarHandle As Long) As Ext
Declare Function ucGetVariableValueStrE Lib "ucalcfmp.dll" Alias "ucGetVariableValueStrE" (ByVal VarHandle As Long) As String
Declare Function ucImagPartE Lib "ucalcfmp.dll" Alias "ucImagPartE" (ByVal t As Dword) As Ext
Declare Function ucLowerBoundE Lib "ucalcfmp.dll" Alias "ucLowerBoundE" (ByVal VarHandle As Long, ByVal Dimension As Long) As Long
Declare Function ucNewThreadE Lib "ucalcfmp.dll" Alias "ucNewThreadE" () As Long
Declare Function ucNewSymbolHandleE Lib "ucalcfmp.dll" Alias "ucNewSymbolHandleE" (t As ucThreadType) As Long
Declare Function ucOpE Lib "ucalcfmp.dll" Alias "ucOpE" (ByVal OpID As Long) As Long
Declare Function ucParamHandleE Lib "ucalcfmp.dll" Alias "ucParamHandleE" (ByVal pIndex As Long, t As ucThreadType) As Long
Declare Function ucParamTypeE Lib "ucalcfmp.dll" Alias "ucParamTypeE" (ByVal pIndex As Long, t As ucThreadType) As Long
Declare Function ucParseE Lib "ucalcfmp.dll" Alias "ucParseE" (Expression As Asciiz, LocalVars As Asciiz, ByVal ExprType As Long, ByVal t As Dword) As Dword
Declare Function ucPrecedenceE Lib "ucalcfmp.dll" Alias "ucPrecedenceE" (Op As Asciiz, ByVal OpCategory As Long, ByVal t As Dword) As Long
Declare Function ucSetCallbackE Lib "ucalcfmp.dll" Alias "ucSetCallbackE" (ByVal FunctionAddress As Dword, ByVal TreeOption As Long, ByVal t As Dword) As Long
Declare Function ucSymbolHandleE Lib "ucalcfmp.dll" Alias "ucSymbolHandleE" (ucSymbol As Asciiz, ByVal SymbolProperty As Long, ByVal t As Dword) As Long
Declare Function ucThreadAddrE Lib "ucalcfmp.dll" Alias "ucThreadAddrE" (ByVal t As Long) As Dword
Declare Function ucTreeGetParamE Lib "ucalcfmp.dll" Alias "ucTreeGetParamE" (ByVal pIndex As Long, t As ucThreadType) As String
Declare Function ucTreePropertyE Lib "ucalcfmp.dll" Alias "ucTreePropertyE" (ByVal NodePtr As Dword, ByVal pIndex As Long, ByVal pOption As Long, t As ucThreadType) As Long
Declare Function ucUpperBoundE Lib "ucalcfmp.dll" Alias "ucUpperBoundE" (ByVal VarHandle As Long, ByVal Dimension As Long) As Long

' These 3 functions come from ucalcfmE.dll instead of ucalcfmP.dll
Declare Function ucEvaluate Lib "ucalcfme.dll" Alias "ucEvaluateE" (ByVal ExpressionPtr As Dword) As Ext
Declare Function ucEvaluateLng Lib "ucalcfme.dll" Alias "ucEvaluateLngE" (ByVal ExpressionPtr As Dword) As Long
Declare Function ucEvaluateQwd Lib "ucalcfme.dll" Alias "ucEvaluateQwdE" (ByVal ExpressionPtr As Dword) As Quad
Declare Function ucEvaluateFxt Lib "ucalcfme.dll" Alias "ucEvaluateFxtE" (ByVal ExpressionPtr As Dword) As Currency
Declare Sub ucEvaluateInitE Lib "ucalcfme.dll" Alias "ucEvaluateInitE" ()
Declare Sub ucReleaseExprPtrE Lib "ucalcfme.dll" Alias "ucReleaseExprPtrE" (ByVal ExpressionPtr As Dword)

Function ucOp(ByVal OpID As Long) As Long
   ucOp = ucOpE(OpID)
End Function

Function ucThreadAddr(ByVal t As Long) As Dword
   If ucInitialized = %ucFalse Then Call ucEvaluateInit
   ucThreadAddr = ucThreadAddrE(t)
End Function

Function ucAlias(AssignFrom$, AssignTo$, Optional ByVal aProperties As Long, Optional ByVal Precedence As Long, Optional ByVal t As Long) As Long
   ucAlias = ucAliasE(AssignFrom$+Chr$(0), AssignTo$+Chr$(0), aProperties, Precedence, ucThreadAddr(t))
End Function

Sub ucDetachSymbol(Symbol$, t As ucThreadType)
   ucDetachSymbolE Symbol$+Chr$(0), t
End Sub

Sub ucErrorClear(Optional ByVal t As Long)
   ucErrorClearE ucThreadAddr(t)
End Sub

Sub ucReleaseExprPtr(ByVal ExpressionPtr As Dword)
   ucReleaseExprPtrE ExpressionPtr
End Sub

Sub ucReleaseItem(ByVal ItemHandle As Long)
   ucReleaseItemE ItemHandle
End Sub

Sub ucReleaseThread(ByVal ucThread As Long)
   If ucThread > 0 Then ucReleaseThreadE ucThread
End Sub

Sub ucSetVariableValue(ByVal VarHandle As Long, ByVal vValue As Ext)
   ucSetVariableValueE VarHandle, vValue
End Sub

Sub ucSetVariableValueStr(ByVal VarHandle As Long, vValue$)
   ucSetVariableValueStrE VarHandle, vValue$+Chr$(0)
End Sub

Function ucArrayInfo(ByVal VarHandle As Long, ByVal InfoOption As Long) As Dword
   ucArrayInfo = ucArrayInfoE(VarHandle, InfoOption)
End Function

Function ucDefine(Definition$, Optional ByVal t As Long) As Long
   ucDefine = ucDefineE(Definition$+Chr$(0), ucThreadAddr(t))
End Function

Function ucDefineConstant(ConstantDef$, Optional ByVal vProperties As Long, Optional ByVal t As Long) As Long
   ucDefineConstant = ucDefineVariableE(ConstantDef$+Chr$(0), 0, 0, vProperties+%ucConstant, t0.ucProperty.DefaultNumericType, ucThreadAddr(t))
End Function

Function ucDefineFunction(FunctionDef$, Optional ByVal FunctionAddress As Dword, Optional ByVal fProperties As Long, Optional ByVal vType As Long, Optional ByVal t As Long) As Long
   ucDefineFunction = ucDefineFunctionE(FunctionDef$+Chr$(0), FunctionAddress, 0, fProperties, vType, ucThreadAddr(t))
End Function

Function ucDefineMacro(Definition$, Optional ByVal t As Long) As Long
   ucDefineMacro = ucDefineE("Macro "+Definition$, ucThreadAddr(t))
End Function

Function ucDefineOperator(OperatorDef$, Optional ByVal Precedence As Long, Optional ByVal OperatorAddress As Dword, Optional ByVal fProperties As Long, Optional ByVal vType As Long, Optional ByVal t As Long) As Long
   ucDefineOperator = ucDefineFunctionE(OperatorDef$+Chr$(0), OperatorAddress, Precedence, fProperties, vType, ucThreadAddr(t))
End Function

Function ucDefineSyntax(Definition$, Optional ByVal t As Long) As Long
   ucDefineSyntax = ucDefineE("Syntax "+Definition$, ucThreadAddr(t))
End Function

Function ucDefineVariable(VariableDef$, Optional ByVal vProperties As Long, Optional ByVal t As Long) As Long
   ' +++ vType should be independent in each thread (update other includes too)
   ucDefineVariable = ucDefineVariableE(VariableDef$+Chr$(0), 0, 0, vProperties, t0.ucProperty.DefaultNumericType, ucThreadAddr(t))
End Function

Function ucAttachVariable(VariableDef$, ByRef vPtr As Dword, Optional ByVal ExprPtr As Dword, Optional ByVal vProperties As Long, Optional ByVal t As Long) As Long
   ucAttachVariable = ucDefineVariableE(VariableDef$+Chr$(0), vPtr, ExprPtr, vProperties, t0.ucProperty.DefaultNumericType, ucThreadAddr(t))
End Function

Function ucErrorHandler(ByVal FunctionAddress As Dword, Optional ByVal t As Long) As Long
   ucErrorHandler = ucSetCallbackE(FunctionAddress, %ucErrorCallback, ucThreadAddr(t))
End Function

Function ucEval(Expression$, Optional ByVal t As Long) As Ext
   ucEval = ucEvalE(Expression$+Chr$(0), ucThreadAddr(t))
End Function

Function ucEvalStr(Expression$, Optional ByVal t As Long) As String
   ucEvalStr = ucEvalStrE(Expression$+Chr$(0), ucThreadAddr(t))
End Function

'Macro ucEvaluate(ExpressionPtr) = ucEvaluateE(ExpressionPtr)

'Function ucEvaluate(ByVal ExpressionPtr As Dword) As Ext
'   ucEvaluate = ucEvaluateE(ExpressionPtr)
'End Function

Function ucEvaluateStr(ByVal ExpressionPtr As Dword) As String
   ucEvaluateStr = ucEvaluateStrE(ExpressionPtr)
End Function

Function ucError() As Long
   ucError = t0.ErrorNumber
End Function

Function ucGetErrorData(ByVal errIndex As Long, Optional ByVal t As Long) As String
   ucGetErrorData = ucGetErrorDataE(errIndex, ucThreadAddr(t))
End Function

Function ucGetVariableValue(ByVal VarHandle As Long) As Ext
   ucGetVariableValue = ucGetVariableValueE(VarHandle)
End Function

Function ucGetVariableValueStr(ByVal VarHandle As Long) As String
   ucGetVariableValueStr = ucGetVariableValueStrE(VarHandle)
End Function

Function ucImagPart(Optional ByVal t As Long) As Ext
   ucImagPart = ucImagPartE(ucThreadAddr(t))
End Function

Function ucLowerBound(ByVal VarHandle As Long, Optional ByVal Dimension As Long) As Long
   If Dimension = 0 Then Dimension = 1
   ucLowerBound = ucLowerBoundE(VarHandle, Dimension)
End Function

Function ucNewThread() As Long
   ucNewThread = ucNewThreadE
End Function

Function ucNewSymbolHandle(t As ucThreadType) As Long
   ucNewSymbolHandle = ucNewSymbolHandleE(t)
End Function

Function ucParse(Expression$, Optional ByVal LocalVars$, Optional ByVal ExprType As Long, Optional ByVal t As Long) As Dword
   ucParse = ucParseE(Expression$+Chr$(0), LocalVars$+Chr$(0), ExprType, ucThreadAddr(t))
End Function

Function ucPrecedence(Op$, Optional ByVal OpCategory As Long, Optional ByVal t As Long) As Long
   ucPrecedence = ucPrecedenceE(Op$+Chr$(0), OpCategory, ucThreadAddr(t))
End Function

Function ucSymbolHandle(ByVal ucSymbol$, Optional ByVal SymbolProperty As Long, Optional ByVal t As Long) As Long
   ucSymbolHandle = ucSymbolHandleE(ucSymbol$+Chr$(0), SymbolProperty, ucThreadAddr(t))
End Function

Function ucErrorData(ByVal errIndex As Long, Optional ByVal t As Long) As String
   ucErrorData = ucGetErrorDataE(errIndex, ucThreadAddr(t))
End Function

Function ucUpperBound(ByVal VarHandle As Long, Optional ByVal Dimension As Long) As Long
   If Dimension = 0 Then Dimension = 1
   ucUpperBound = ucUpperBoundE(VarHandle, Dimension)
End Function

Function ucVariableValue(ByVal VarHandle As Long) As Ext
   ucVariableValue = ucGetVariableValueE(VarHandle)
End Function

Function ucVariableValueStr(ByVal VarHandle As Long) As String
   ucVariableValueStr = ucGetVariableValueStrE(VarHandle)
End Function

Function ucGetThreadInfo(ByVal ucThread As Long, t As ucThreadType) As Dword
   ucGetThreadInfo = ucGetThreadInfoE(ucThread, t)
End Function
Sub ucSetThreadAddr(ByVal ucThread As Long, t As ucThreadType)
   ucSetThreadAddrE ucThread, t
End Sub

Sub ucSetTrigMode(ByVal TrigMode As Long, t As ucThreadType)
   ucSetTrigMode TrigMode, t
End Sub

Sub ucReturn(ByVal vValue As Ext, t As ucThreadType)
   ucReturnE vValue, t
End Sub

' +++ Add all other new exported routines and for other include files
Sub ucReturnCplx(ByVal vValue As Ext, t As ucThreadType)
   ucReturnCplxE vValue, t
End Sub

Sub ucReturnStr(Expression$, t As ucThreadType)
   ucReturnStrE Expression$+Chr$(0), t
End Sub

Function ucGetParamStr(ByVal ParamIndex As Long, t As ucThreadType) As String
   ucGetParamStr = ucGetParamStrE(ParamIndex, t)
End Function
Sub ucSetParamStr(ByVal ParamIndex As Long, Argument$, t As ucThreadType)
   ucSetParamStrE ParamIndex, Argument$+Chr$(0), t
End Sub

Function ucParamStr(ByVal ParamIndex As Long, t As ucThreadType) As String
   ucParamStr = ucGetParamStrE(ParamIndex, t)
End Function

Sub ucLicense(SerialNumber$)
   ucLicenseE SerialNumber$+Chr$(0)
End Sub

Function ucElementCategory(t As ucThreadType) As Long
   ucElementCategory = ucTreePropertyE(0, 0, %ucElementCat, t)
End Function

Function ucTreeProperty(ByVal NodePtr As Dword, ByVal pIndex As Long, ByVal pOption As Long, t As ucThreadType) As Long
   ucTreeProperty = ucTreePropertyE(NodePtr, pIndex, pOption, t)
End Function

Function ucTreeNodeProperty(ByVal pOption As Long, t As ucThreadType) As Long
   ucTreeNodeProperty = ucTreePropertyE(0, 0, pOption, t)
End Function

Function ucTreeParamProperty(ByVal pIndex As Long, ByVal pOption As Long, t As ucThreadType) As Long
   ucTreeParamProperty = ucTreePropertyE(0, pIndex, pOption, t)
End Function

Function ucTreeParentProperty(ByVal pOption As Long, t As ucThreadType) As Long
   Dim ParentPtr As Dword

   ParentPtr = ucTreePropertyE(0, 0, %ucParentPtr, t)
   ucTreeParentProperty = ucTreePropertyE(ParentPtr, 0, pOption, t)
End Function

Function ucPreParse(ByVal FunctionAddress As Dword, Optional ByVal t As Long) As Long
   ucPreParse = ucSetCallbackE(FunctionAddress, %ucTreePre, ucThreadAddr(t))
End Function

Function ucTreeParseDownEarly(ByVal FunctionAddress As Dword, Optional ByVal t As Long) As Long
   ucTreeParseDownEarly = ucSetCallbackE(FunctionAddress, %ucTreeDownEarly, ucThreadAddr(t))
End Function

Function ucTreeParseDown(ByVal FunctionAddress As Dword, Optional ByVal t As Long) As Long
   ucTreeParseDown = ucSetCallbackE(FunctionAddress, %ucTreeDown, ucThreadAddr(t))
End Function

Function ucTreeParseUp(ByVal FunctionAddress As Dword, Optional ByVal t As Long) As Long
   ucTreeParseUp = ucSetCallbackE(FunctionAddress, %ucTreeUp, ucThreadAddr(t))
End Function

Function ucGetExpr(t As ucThreadType) As String
   ucGetExpr = ucTreeGetParamE(-1, t)
End Function

Function ucTreeGetExpr(t As ucThreadType) As String
   ucTreeGetExpr = ucTreeGetParamE(-1, t)
End Function

Function ucTreeGetFunc(t As ucThreadType) As String
   ucTreeGetFunc = ucTreeGetParamE(0, t)
End Function

Function ucTreeGetOp(t As ucThreadType) As String
   ucTreeGetOp = ucTreeGetParamE(0, t)
End Function

Function ucTreeGetParam(ByVal pIndex As Long, t As ucThreadType) As String
   ucTreeGetParam = ucTreeGetParamE(pIndex, t)
End Function

Function ucTreeGetParamList(t As ucThreadType) As String
   ucTreeGetParamList = ucTreeGetParamE(-2, t)
End Function

Sub ucSetExpr(Expression$, t As ucThreadType)
   ucTreeSetParamE -1, Expression$+Chr$(0), t
End Sub

Sub ucTreeSetExpr(Expression$, t As ucThreadType)
   ucTreeSetParamE -1, Expression$+Chr$(0), t
End Sub

Sub ucTreeSetFunc(TreeFunc$, t As ucThreadType)
   ucTreeSetParamE 0, TreeFunc$+Chr$(0), t
End Sub

Sub ucTreeSetOp(TreeOp$, t As ucThreadType)
   ucTreeSetParamE 0, TreeOp$+Chr$(0), t
End Sub

Sub ucTreeSetParam(ByVal pIndex As Long, pData$, t As ucThreadType)
   ucTreeSetParamE pIndex, pData$+Chr$(0), t
End Sub

Sub ucTreeSetParamList(pData$, t As ucThreadType)
   ucTreeSetParamE -2, pData$+Chr$(0), t
End Sub

Function ucParam(ByVal x As Long, t As ucThreadType) As Ext
   ucParam = ucGetParamE(x, t)
End Function

Function ucParamCplx(ByVal x As Long, t As ucThreadType) As Ext
   ucParamCplx = ucGetParamCplxE(x, t)
End Function

Function ucParamHandle(ByVal pIndex As Long, t As ucThreadType) As Long
   ucParamHandle = ucParamHandleE(pIndex, t)
End Function

Function ucParamType(ByVal pIndex As Long, t As ucThreadType) As Long
   ucParamType = ucParamTypeE(pIndex, t)
End Function

Sub ucSetErrorData(ByVal errIndex As Long, vValue$, t As ucThreadType)
   ucSetErrorDataE errIndex, vValue$+Chr$(0), t
End Sub

Function uc_Precedence(t As ucThreadType) As Ext
   uc_Precedence = ucPrecedence(ucParamStr(1, t), ucParam(2, t), t.ucThread)
End Function

Function uc_PrecedenceMax(t As ucThreadType) As Ext
   uc_PrecedenceMax = t.PrecedenceMax
End Function

Function uc_PrecedenceMin(t As ucThreadType) As Ext
   uc_PrecedenceMin = t.PrecedenceMin
End Function

Function ucEvaluateCplx(ByVal ExpressionPtr As Dword, ucValue As ucComplexNumberType, t As ucThreadType) As Ext
   ucValue.Real = ucEvaluate(ExpressionPtr)
   ucValue.Imag = ucImagPartE(ucThreadAddr(t.ucThread))
   ucEvaluateCplx = ucValue.Real
End Function

#If Not %Def(%ucAddOnLibrary)
#If Not %Def(%ucAltErrMsg)
Function ucErrorMessage(Optional ByVal ErrorNumber As Long, Optional ByVal t As Long) As String
   Dim ErrMsg As String

   If ErrorNumber = 0 Then ErrorNumber = t0.ErrorNumber

   Select Case As Long ErrorNumber
   Case %ucErr_None
      ' No Error
   Case %ucErr_ucErrorData
      ErrMsg = ucErrorData(0, t)
   Case %ucErr_SyntaxError
      ErrMsg = "Syntax error"
   Case %ucErr_MismatchedParenth
      ErrMsg = "Mismatched parenthesis"
   Case %ucErr_MismatchedQuotes
      ErrMsg = "Mismatched quotes"
   Case %ucErr_UndefinedVariable
      ErrMsg = "Undefined variable"
   Case %ucErr_UndefinedFunction
      ErrMsg = "Undefined function"
   Case %ucErr_IncorrectArgType
      ErrMsg = "Incorrect argument type"
   Case %ucErr_InvalidNumberOfArgs
      ErrMsg = "Invalid number of function arguments"
   Case %ucErr_InvalidOperandType
      ErrMsg = "Invalid operand type"
   Case %ucErr_InvalidVariableName
      ErrMsg = "Invalid variable name"
   Case %ucErr_ImplicitMult
      ErrMsg = "Implicit multiplication error"
   Case %ucErr_IncompleteBlock
      ErrMsg = "Incomplete syntax block"

   Case %ucErr_InvalidOperation
      ErrMsg = "Invalid floating point operation"
   Case %ucErr_DenormalOperand
      ErrMsg = "Denormal floating point operand"
   Case %ucErr_DivisionByZero
      ErrMsg = "Division by 0"
   Case %ucErr_Overflow
      ErrMsg = "Floating point overflow"
   Case %ucErr_Underflow
      ErrMsg = "Floating point underflow"
   Case %ucErr_InexactResult
      ErrMsg = "Inexact floating point result"
   Case %ucErr_SubscriptOutOfRange
      ErrMsg = "Array subscript out of range"
   Case %ucErr_StackOverflow
      ErrMsg = "Stack overflow"

   Case %ucErr_OperationDenied
      ErrMsg = "Cannot perform requested operation"
   Case %ucErr_PropertyConflict
      ErrMsg = "Definition property conflict"
   Case %ucErr_ConstantWriteDenied
      ErrMsg = "Assignment to a constant not allowed"
   Case %ucErr_FunctionDefError
      ErrMsg = "Invalid function or operator definition"
   Case %ucErr_FunctionLocked
      ErrMsg = "Function or operator is locked"

   Case %ucErr_DefinitionSpaceFull
      ErrMsg = "Definition space is full"
   Case %ucErr_ThreadInitFailed
      ErrMsg = "Thread initialization failed"
   Case %ucErr_UnsupportedFeature
      ErrMsg = "Requires advanced edition of UCalc FMP"
   Case Else
      ErrMsg = "Unknown error"
      ' numbers up to 500 are reserved for future use by UCalc
      ' Use Case 501 and beyond for user-defined errors
   End Select

   If ErrorNumber > 1 Then ErrMsg = "Error #" + LTrim$(Str$(ErrorNumber)) + ": "  + ErrMsg
   ucErrorMessage = ErrMsg
End Function
#EndIf

Sub ucReset(t As ucThreadType)
   Dim Between_Mult_Pow As Long
   Dim Between_And_Less As Long
   Dim Between_Div_Mod  As Long

   ReDim ucItem(t.ucProperty.DefSpace.BaseDefinitionSpace _
               +t.ucProperty.DefSpace.DefinitionSpace     _
               *t.ucProperty.DefSpace.MaxThreadCount) As ucHandleType

   t.Context = t.Context Or %ucContext_ucReset
   t.Handshake.LanguageID = %uc_Lang_PowerBASIC
   t.Handshake.HandleAddr = VarPtr(ucItem(0))
   ucEvaluateInitE()
   ucResetE t
   t.ucProperty.ComplexNumbers = %ucTrue
   t.ucProperty.EscapeChar     = Asc("~")

   ucDefineConstant "Byte     = uc_Integer8u"
   ucDefineConstant "Integer  = uc_Integer16"
   ucDefineConstant "Word     = uc_Integer16u"
   ucDefineConstant "Long     = uc_Integer32"
   ucDefineConstant "DWord    = uc_Integer32u"
   ucDefineConstant "Quad     = uc_Integer64"
   ucDefineConstant "Currency = uc_Currency"
   ucDefineConstant "CurrencyX= uc_CurrencyExt"
   ucDefineConstant "Single   = uc_FloatSingle"
   ucDefineConstant "Double   = uc_FloatDouble"
   ucDefineConstant "Extended = uc_FloatExtended"
   ucDefineConstant "AsciiZ   = uc_StringLPSTR"
   ucDefineConstant "String   = uc_StringAnsi"
   ucDefineConstant "Ext      = Extended"
   ucDefineConstant "FixedPt  = Currency"
   ucDefineConstant "Any      = uc_Any"

   Between_Mult_Pow = (ucPrecedence("*") + ucPrecedence("^")) / 2
   Between_And_Less = (ucPrecedence("AND") + ucPrecedence("<")) / 2
   Between_Div_Mod  = (ucPrecedence("/") + ucPrecedence("MOD")) / 2

   t.ucProperty.FunctionProperties = %ucFastCallback + %ucOverload

'   ucDefineOperator "*[x As Long, y As Long] As Long", ucPrecedence("*"), ucOp(%Op_Int_Multiply)
   ucDefineOperator "-[x As Long, y As Long] As Long", ucPrecedence("-"), ucOp(%Op_Int_Subtract)
   ucDefineOperator "+[x As Long, y As Long] As Long", ucPrecedence("+"), ucOp(%Op_Int_Add)
   ucDefineOperator "/[x As Long, y As Long] As Extended", ucPrecedence("/"), ucOp(%Op_Int_DivideToExt)
   ucDefineOperator "\[x As Long, y As Long] As Long", Between_Div_Mod, ucOp(%Op_Int_DivideToInt)
   ucDefineOperator "Mod[x As Long, y As Long] As Long", ucPrecedence("MOD"), ucOp(%Op_Int_Mod)
   ucDefineOperator "^[x As Extended, y As Long] As Extended", ucPrecedence("^"), ucOp(%Op_IntPower)

'   ucDefineOperator "*[x As Quad, y As Quad] As Quad", ucPrecedence("*"), ucOp(%Op_Qwd_Multiply)
   ucDefineOperator "-[x As Quad, y As Quad] As Quad", ucPrecedence("-"), ucOp(%Op_Qwd_Subtract)
   ucDefineOperator "+[x As Quad, y As Quad] As Quad", ucPrecedence("+"), ucOp(%Op_Qwd_Add)
   ucDefineOperator "/[x As Quad, y As Quad] As Extended", ucPrecedence("/"), ucOp(%Op_Qwd_DivideToExt)
   ucDefineOperator "\[x As Quad, y As Quad] As Quad", Between_Div_Mod, ucOp(%Op_Qwd_DivideToInt)

'   ucDefineOperator "*[x As FixedPt, y As FixedPt] As FixedPt", ucPrecedence("*"), ucOp(%Op_Cur_Multiply)
   ucDefineOperator "-[x As FixedPt, y As FixedPt] As FixedPt", ucPrecedence("-"), ucOp(%Op_Cur_Subtract)
   ucDefineOperator "+[x As FixedPt, y As FixedPt] As FixedPt", ucPrecedence("+"), ucOp(%Op_Cur_Add)
   ucDefineOperator "/[x As FixedPt, y As FixedPt] As Extended", ucPrecedence("/"), ucOp(%Op_Cur_DivideToExt)

   ucDefineFunction "Fact(x As Long)", ucOp(%Op_Func_Factorial)
   ucDefineFunction "GetFPU(x As Long) As Long", ucOp(%Op_Func_GetFPU)
   ucDefineFunction "SetFPU(x As Long)", ucOp(%Op_Func_SetFPU)
   ucDefineFunction "SetOrFPU(x As Long)", ucOp(%Op_Func_SetOrFPU)
   ucDefineFunction "SetXorFPU(x As Long)", ucOp(%Op_Func_SetXorFPU)
   ucDefineFunction "RestoreFPU(x As Long)", ucOp(%Op_Func_RestoreFPU)
   ucDefineFunction "Ceil(x)", ucOp(%Op_Func_Ceil)
   ucDefineFunction "Not(x)", ucOp(%Op_Func_Not)

   ' String relational (comparison) operators
   t.ucProperty.DefaultPrecedence = ucPrecedence("<")
   t.ucProperty.FunctionProperties = %ucDynamic + %ucLocked + %ucOverride + %ucOverload
   ucDefineFunction "StrCmp(a As String, b As String) = StrCmp(0, 0, a, b)"
   ucDefineOperator " =[a As String, b As String] = StrCmp(a, b) = 2"
   ucDefineOperator "<>[a As String, b As String] = StrCmp(a, b) <> 2"
   ucDefineOperator " >[a As String, b As String] = StrCmp(a, b) = 3"
   ucDefineOperator ">=[a As String, b As String] = StrCmp(a, b) > 1"
   ucDefineOperator " <[a As String, b As String] = StrCmp(a, b) = 1"
   ucDefineOperator "<=[a As String, b As String] = StrCmp(a, b) < 3"

   t.ucProperty.FunctionProperties = %ucInline + %ucLocked + %ucOverride + %ucOverload

   ucDefineFunction "Precedence(Operator$, OpCategory = 0)", CodePtr(uc_Precedence)
   ucDefineFunction "PrecedenceMax()", CodePtr(uc_PrecedenceMax)
   ucDefineFunction "PrecedenceMin()", CodePtr(uc_PrecedenceMin)

   ucDefineMacro ".[x]  ::= (0.x) ::: Precedence = PrecedenceMax"
   ucDefineMacro " [x]. ::= (x.0) ::: Precedence = PrecedenceMax"

   ' Hex, binary, octal notations.  Includes support for non-integers
   ucDefineSyntax "#b (Number:'[0-1]+') [(Frac:'\.[0-1]+'=)] ::= BaseConvert(~qNumber~jFrac~q, 2) ::: Precedence = PrecedenceMax"
   ucDefineSyntax "#o (Number:'[0-7]+') [(Frac:'\.[0-7]+'=)] ::= BaseConvert(~qNumber~jFrac~q, 8) ::: Precedence = PrecedenceMax"
   ucDefineSyntax "#h (Number:'[0-9A-F]+') [(Frac:'\.[0-9A-F]+'=)] ::= BaseConvert(~qNumber~jFrac~q, 16) ::: Precedence = PrecedenceMax"
   ucDefineSyntax "#i ::= ~i"

   ' Alternative notations
'   ucDefineOperator "  &[a,b] = a and b", ucPrecedence("and")
'   ucDefineOperator "  |[a,b] = a or  b", ucPrecedence("or ")
'   ucDefineOperator " =<[a,b] = a <=  b", ucPrecedence("<= ")
'   ucDefineOperator " =>[a,b] = a >=  b", ucPrecedence(">= ")
   ucDefineOperator " ==[a,b] = a  =  b", ucPrecedence("=")
   ucDefineOperator "DIV[a,b] = a  \  b", ucPrecedence("\")
   ucDefineOperator " **[a,b] = a  ^  b", ucPrecedence("^")
   ucDefineMacro    "  &[TextA,TextB] ::= TextA + TextB"
   ucDefineOperator "==[TextA$,TextB$] = TextA = TextB", ucPrecedence("=")
   ucDefineOperator "Imp[a, b] = (Not(a)) Or b", ucPrecedence("OR")-1
   ucDefineOperator "Eqv[a, b] = (a and b) Or (not(a) and not(b))", ucPrecedence("OR")-1
   ucDefineFunction "Concatenate$(a$,b$) = a + b"
   ucDefineFunction "Sign(x) = Sgn(x)"
   ucDefineFunction "Sqrt(x) = Sqr(x)"
   ucDefineFunction "Ln(x)   = Log(x)"
   ucDefineFunction "Atan(x) = Atn(x)" ' +++ should do these after complex defs

   ucDefineOperator "[x As Long]! = fact(x)", t.PrecedenceMax
   ucDefineOperator "[x]% = x/100", t.PrecedenceMax
   ucDefineOperator "%[x,y] = x mod y", ucPrecedence("MOD")
   ucDefineOperator "-[x] = 0-x", Between_Mult_Pow
   ucDefineOperator "+[x] = 0+x", Between_Mult_Pow
   ucDefineOperator "Not[x] = Not(x)", Between_And_Less

   ucDefineFunction "GetFPU() = GetFPU(0)"
   ucDefineFunction "RestoreFPU() = RestoreFPU(0)"
   ucAlias "GetFPU", "GetFPU(0)"
   ucAlias "RestoreFPU", "RestoreFPU(0)"
   ucAlias "SetFPU(setting, Expression)", "(SetFPU(setting)+(Expression)+RestoreFPU(0))"
   ucAlias "SetOrFPU(setting, Expression)", "(SetOrFPU(setting)+(Expression)+RestoreFPU(0))"
   ucAlias "SetXorFPU(setting, Expression)", "(SetXorFPU(setting)+(Expression)+RestoreFPU(0))"

   ucDefineFunction "Inf() = SetOrFPU(4, 1/0)", 0, %ucDynamic
   ucDefineFunction "NaN() = SetOrFPU(1, 0/0)", 0, %ucDynamic

   ' Hyperbolic functions
   ucDefineFunction "Sinh(x)  = (exp(x) - exp(-x)) / 2"
   ucDefineFunction "Cosh(x)  = (exp(x) + exp(-x)) / 2"
   ucDefineFunction "Tanh(x)  = (exp(2x) - 1) / (exp(2x) + 1)"
   ucDefineFunction "Coth(x)  = (exp(2x) + 1) / (exp(2x) - 1)"
   ucDefineFunction "Sech(x)  = 2 / (exp(x) + exp(-x))"
   ucDefineFunction "Csch(x)  = 2 / (exp(x) - exp(-x))"
   ucDefineFunction "Asinh(x) = Log(x + Sqr(x ^ 2 + 1))"
   ucDefineFunction "Acosh(x) = Log(x + Sqr(x ^ 2 - 1))"
   ucDefineFunction "Atanh(x) = Log((1 + x) / (1 - x)) / 2"
   ucDefineFunction "Acoth(x) = Log((x + 1) / (x - 1)) / 2"
   ucDefineFunction "Asech(x) = Log((1 + Sqr(1 - x ^ 2)) / x)"
   ucDefineFunction "Acsch(x) = Log(1 / x + Sqr(1 + x ^ 2) / Abs(x))"

   ' A few constants needed in some definitions further down
   ucDefineConstant "Pi     = Atn(1) * 4"
   ucDefineConstant "HalfPI = Pi/2"

   ' Other functions
   ucDefineFunction "Sec(x)   = 1 / Cos(x)"
   ucDefineFunction "Csc(x)   = 1 / Sin(x)"
   ucDefineFunction "Cot(x)   = 1 / Tan(x)"
   ucDefineFunction "Asin(x)  = SetOrFPU(4, Atn(x / Sqr(1 - x * x)))" ' +++ make provisions for complex
   ucDefineFunction "Acos(x)  = HalfPI - Asin(x)"
   ucDefineFunction "Log2(x)  = Log(x) / Log(2)"
   ucDefineFunction "Exp2(x)  = 2 ^ x"
   ucDefineFunction "Log10(x) = Log(x) / Log(10)"
   ucDefineFunction "Exp10(x) = 10 ^ x"

   ' Complex operators & functions
   t.ucProperty.FunctionProperties = %ucDynamic + %ucLocked + %ucOverride + %ucOverload + %ucComplex

   ucDefineOperator "+[x] = x.Real + x.Imag*#i", Between_Mult_Pow
   ucDefineOperator "-[x] = (-x.Real) + (-x.Imag)*#i", Between_Mult_Pow

   ucDefineOperator "+[x,y] = x.Real+y.Real + (x.Imag+y.Imag)*#i", ucPrecedence("+")
   ucDefineOperator "*[x,y] = (x.Real*y.Real - x.Imag*y.Imag)+(x.Real*y.Imag + x.Imag*y.Real)*#i", ucPrecedence("*")
   ucDefineOperator "-[x,y] = (x.Real-y.Real) + (x.Imag-y.Imag)*#i", ucPrecedence("-")
   ucDefineOperator "/[x,y] = x.Real*y.Real/(y.Real^2+y.Imag^2) + x.Imag*y.Imag" _
   + "/(y.Real^2+y.Imag^2) + (x.Imag*y.Real/(y.Real^2+y.Imag^2)-x.Real*y.Imag" _
   + "/(y.Real^2+y.Imag^2))*#i", ucPrecedence("/")

   ucDefineFunction "Abs(x) = Sqr(x.Real^2 + x.Imag^2) + 0*#i"
   ucDefineFunction "Sgn(x) = x.Real/Abs(x) + x.Imag/Abs(x)*#i"
   ucDefineFunction "Sin(x) = Sin(x.Real)*Cosh(x.Imag) + Cos(x.Real)*Sinh(x.Imag)*#i"
   ucDefineFunction "Cos(x) = Cos(x.Real)*Cosh(x.Imag) - Sin(x.Real)*Sinh(x.Imag)*#i"
   ucDefineFunction "Phase(x) = iif(x.Imag=0, iif(x.Real>0, 0, pi), Atn(x.Imag/x.Real)) + 0*#i"
   ucDefineFunction "Conj(x)  = x.Real - x.Imag*#i"
   ucDefineFunction "Exp(x)   = Exp(x.Real)*Cos(x.Imag) + Exp(x.Real)*Sin(x.Imag)*#i"
   ucDefineFunction "Imag(x)  = x.Imag + 0*#i"
   ucDefineFunction "Real(x)  = x.Real + 0*#i"

   ucDefineFunction "Sqr(x)   = Sqr(Sqr(x.Real^2 + x.Imag^2)+x.Real)/Sqr(2)" _
   + "+ (Sqr(Sqr(x.Real^2 + x.Imag^2)-x.Real)*iif(x.Imag < 0, -1, 1)/Sqr(2))*#i"
   ucDefineFunction "Sqrt(x)  = Sqr(x.Real+x.Imag*#i)"

   ucDefineFunction "Log(x) = iif(x.Real=0 and x.Imag=0, -Inf," _
   + "Log(x.Real^2 + x.Imag^2)/2)" _
   + "+iif(x.Real=0 and x.Imag=0, 0, iif(x.Real>0 and x.Imag=0, 0," _
   + "iif(x.Imag=0, pi, pi*sgn(x.Imag)/2-Atn(x.Real/x.Imag))))*#i"

   ucDefineFunction "_ZeroToZero() = 1", 0, %ucDynamic+%ucOverwrite
'  ucDefineFunction "_ZeroToZero() = 0", 0, %ucDynamic+%ucOverwrite
'  ucDefineFunction "_ZeroToZero() = NaN", 0, %ucDynamic+%ucOverwrite
'  ucDefineFunction "_ZeroToZero() = SetError(1, '0^0 Error')", 0, %ucDynamic+%ucOverwrite

   ucDefineOperator "^[x,y] = iif(x.Imag = 0, iif(x.Real > 0," _
   + "x.Real^y.Real*Cos(y.Imag*Log(x.Real))," _
   + "iif(x.Real=0 and y.Real=0 and y.Imag=0, _ZeroToZero()," _
   + "(-x.Real)^y.Real*COS(y.Imag*LOG(-x.Real)+pi*y.Real)/EXP(pi*y.Imag)))," _
   + "EXP(y.Imag*ATN(x.Real/x.Imag))*SQR((x.Imag^2+x.Real^2)^y.Real)" _
   + "*COS(pi*y.Real*sgn(x.Imag)/2-y.Real*ATN(x.Real/x.Imag)+y.Imag" _
   + "*LOG(x.Imag^2+x.Real^2)/2)/SQR(EXP(pi*y.Imag*Sgn(x.Imag))))-55+55" _
   + "+iif(x.Imag = 0, iif(x.Real > 0," _
   + "x.Real^y.Real*SIN(y.Imag*LOG(x.Real))," _
   + "iif(x.Real=0 and y.Real=0 and y.Imag=0, 0," _
   + "(-x.Real)^y.Real*SIN(y.Imag*LOG(-x.Real)+pi*y.Real)/EXP(pi*y.Imag)))," _
   + "EXP(y.Imag*ATN(x.Real/x.Imag))*SQR((x.Imag^2+x.Real^2)^y.Real)" _
   + "*SIN(pi*y.Real*sgn(x.Imag)/2-y.Real*ATN(x.Real/x.Imag)+y.Imag" _
   + "*LOG(x.Imag^2+x.Real^2)/2)/SQR(EXP(pi*y.Imag*Sgn(x.Imag)))-55+55)*#i", _
     ucPrecedence("^")

   t.ucProperty.FunctionProperties = %ucDynamic + %ucOverload + %ucOverwrite + %ucComplex + %ucRecursion
   t.Context = t.Context Xor %ucContext_ucReset
   t.ucProperty.DefaultPrecedence = Between_Div_Mod ' arbitrarily chosen
End Sub


'Sub Class_Initialize()

Sub ucEvaluateInit()
   ucInitialized = %ucTrue
   ucGetThreadInfo 0, t0
   ucSetThreadAddr 0, t0
   ucReset t0
'  ucNewThread

'  ucThread = ucNewThread()
'  If ucThread = -1 Then Me = Nothing
End Sub

#Else

Sub ucEvaluateInit()
   Dim t As ucThreadType

   ucInitialized = %ucTrue
   ucGetThreadInfo 0, t
   ReDim ucItem(t.ucProperty.DefSpace.BaseDefinitionSpace _
               +t.ucProperty.DefSpace.DefinitionSpace     _
               *t.ucProperty.DefSpace.MaxThreadCount)     _
               As ucHandleType At t.Handshake.HandleAddr
' +++ Adjust this in other includes.
End Sub

#EndIf
