Attribute VB_Name = "UCalc"
' -------------------------------------------------------------------------
' ucalcvb.bas : Module file
' Designed to work with Visual Basic 5.0 or above.
'
' UCalc Fast Math Parser (Beta N)
' Copyright (c) 1998-2005 by Daniel Corbier.  All rights reserved.
' http://www.ucalc.com/mathparser
' -------------------------------------------------------------------------
#Const ucLibraryAddon = False
Option Explicit

' +++ VB include needs major revision
Type ucDefinitionSpaceType
   BaseDefinitionSpace   As Long
   DefinitionSpace       As Long
   InlineBuffer          As Long
   MaxErrorData          As Long
   MaxThreadCount        As Long
End Type

Type ucForInternalUseType
   LanguageID            As Long
   ucOperator            As Long
   ucEvalAddr            As Long
   ucEvaluateAddr        As Long
   ucThreadAddr          As Long
   ParamListAddr         As Long
   HandleAddr            As Long
   ReleaseExprPtr        As Long
   TreeIndex             As Long
   TreeParamInfoAddress  As Long
   HaltEvaluation        As Long
   ThreadStatus          As Long
   PrevOperationCount    As Long
   PrevItemCount         As Long
   CompareOld            As Long
   IsStringValue         As Long
   ucStack               As Long
   Infinity              As Long
   TrigFactor            As Long
   TrigFactorInv         As Long
   
   DataGroup1            As Long
   DataGroup1b           As Long
   DataGroup1c           As Long
   DataGroup2            As Long
   DataGroup2b           As Long
   DataGroup2c           As Long
   DataGroup3            As Long
   DataGroup3b           As Long
   DataGroup3c           As Long
   DataGroup4            As Long
   DataGroup4b           As Long
   DataGroup4c           As Long
   StringPtr1            As Long
End Type

Type ucPropertyType
   DefaultNumericType          As Long
   DefaultExternalStringType   As Long
   FPU_ControlWord_UCalc       As Long
   FPU_ControlWord_HostProgram As Long
   FPU_Turbo                   As Long
   CaseSensitive               As Long
   ComplexNumbers              As Long
   ImagSymbol                  As Long
   RelationalTrue              As Long
   ImplicitMultiply            As Long
   MaxParamPerFunction         As Long
   ArrayProperties             As Long
   FunctionProperties          As Long
   StackSize                   As Long
   DefaultPrecedence           As Long
   OutputDigits                As Long
   QuoteChar                   As Long
   QuoteCharSynonym            As Long
   ArgumentSeparator           As Long
   DecimalSeparator            As Long
   EscapeChar                  As Long
   DefSpace                    As ucDefinitionSpaceType
End Type

Type ucHandleType
   ucThread              As Long
   Info                  As Long
   DataType              As Long
   DataPtr               As Long
   DataCplxPtr           As Long
   SymbolName            As Long
   SortedHandle          As Long
   OverloadHandle        As Long
   OpPrecedence          As Long
   ParamCountHiLo        As Long ' Lo = ReqParamCount; Hi = ParamCount
   CallbackPtr           As Long
   FuncArgs              As Long
   FuncArgHandleData     As Long
   FuncArgTypeList       As Long
   VarValueStr           As Long
   
   DataGroup1            As Long
   DataGroup1b           As Long
   DataGroup1c           As Long
   DataGroup2            As Long
   DataGroup2b           As Long
   DataGroup2c           As Long
   StringPtr1            As Long
   StringPtr2            As Long
   StringPtr3            As Long
   StringPtr4            As Long
   StringPtr5            As Long
   
   UserData1             As Long
   UserData2             As Long
   UserData3             As Long
   UserData4             As Long
End Type

Type ucThreadType
   ucThread              As Long
   Context               As Long
   ParamCount            As Long
   ErrorNumber           As Long
   TreeDirection         As Long
   DefSpace              As Long
   OperationCount        As Long
   ItemCount             As Long
   Sequence              As Long
   SequenceLast          As Long
   ExprHasChanged        As Long
   IsFunctionParse       As Long
   ComplexStatus         As Long
   ucParamAddr           As Long
   ucParamHandleAddr     As Long
   SymbolStart           As Long
   SymbolEnd             As Long
   PrecedenceMax         As Long
   PrecedenceMin         As Long
   VarHandle             As Long
   OpHandle              As Long
   StringWasReturned     As Long
   ReturnString          As Long
   ReturnValue           As Long
   ReturnComplex         As Long
   ucProperty            As ucPropertyType
   Handshake             As ucForInternalUseType
   User1                 As Long
   User2                 As Long
   User3                 As Long
   User4                 As Long
End Type

Public Type ucComplexNumberType
   Real As Double
   Imag As Double
End Type

Public Const ucTrue = 1
Public Const ucFalse = 0

' LanguageID
Public Const uc_Lang_Unsupported = 0
Public Const uc_Lang_Debug = 1
Public Const uc_Lang_PowerBasic = 2
Public Const uc_Lang_Delphi = 3
Public Const uc_Lang_VisualBasic = 4
Public Const uc_Lang_VisualBasicNET = 5
Public Const uc_Lang_CPlusPlus = 6

' Type constants for vType arg of ucDefineFunction and ucDefineVariable
'[[+++ Revise this list later]]
Public Const ucType_Integer8 = 1
Public Const ucType_Integer8u = 2
Public Const ucType_Integer16 = 3
Public Const ucType_Integer16u = 4
Public Const ucType_Integer32 = 5
Public Const ucType_Integer32u = 6
Public Const ucType_Integer64 = 7
Public Const ucType_Integer64u = 8
Public Const ucType_Currency = 9
Public Const ucType_CurrencyExt = 10
Public Const ucType_FloatSingle = 11
Public Const ucType_FloatDouble = 12
Public Const ucType_FloatExtended = 13
Public Const ucType_StringBSTR = 14     ' For VB
Public Const ucType_StringLPSTR = 15    ' (ASCIIZ) C++ char default
Public Const ucType_StringAnsi = 16     ' PowerBASIC and Delphi default
Public Const ucType_Any = 17

Public Const ucType_Long = ucType_Integer32
Public Const ucType_Double = ucType_FloatDouble
Public Const ucType_Extended = ucType_FloatExtended
Public Const ucType_FixedPt = ucType_Currency
Public Const ucType_AnsiString = ucType_StringAnsi

' Context options (t.Context)
Public Const ucContext_Escape = 1
Public Const ucContext_Try = 2
Public Const ucContext_ucDefine = 4
Public Const ucContext_ucDefineFunction = 8
Public Const ucContext_ucDefineVariable = 16
Public Const ucContext_ucEval = 32
Public Const ucContext_ucEvalStr = 64
Public Const ucContext_ucEvaluateStr = 128
Public Const ucContext_ucParse = 256
Public Const ucContext_ucParseOrucEval = 512
Public Const ucContext_ucReset = 1024
Public Const ucContext_Reserved = 2048

' Options for fProperties parameter of ucDefineFunction
' and for ucProperty.FunctionProperties
Public Const ucLeftToRight = 0
Public Const ucInline = 1
Public Const ucDynamic = 2
Public Const ucFastCallback = 4
Public Const ucLocal = 8
Public Const ucComplex = 16
Public Const ucLocked = 32
Public Const ucConstant = ucLocked
Public Const ucOverride = 64
Public Const ucOverwrite = 128
Public Const ucOverload = 256
Public Const ucRecursion = 512
Public Const ucBootstrap = 1024
Public Const ucDelaySyntax = 1024
Public Const ucRightToLeft = 2048

' For ucTreeProperty, ucTreeNodeProperty,
' ucTreeParamProperty, and ucTreeParentProperty
Public Const ucIsInline = 0
Public Const ucIsDynamic = 1
Public Const ucIsFastCallback = 2
Public Const ucIsLocal = 3
Public Const ucIsComplex = 4
Public Const ucIsLocked = 5
Public Const ucIsConstant = ucIsLocked
Public Const ucIsOverride = 6
Public Const ucIsOverwrite = 7
Public Const ucIsOverload = 8
Public Const ucIsRecursion = 9
Public Const ucIsBootstrap = 10
Public Const ucIsDelaySyntax = 10
Public Const ucIsRightToLeft = 11

Public Const ucIsArray = 12

Public Const ucIsFunction = 13
Public Const ucIsOperator = 14
Public Const ucIsInfixOp = 15
Public Const ucIsPostfixOp = 16
Public Const ucIsPrefixOp = 17
Public Const ucIsVariable = 18
Public Const ucIsJITVariable = 19
Public Const ucIsJITExpression = 20
Public Const ucIsString = 21
Public Const ucIsExternal = 22
Public Const ucIsActive = 23
Public Const ucIsItemCount = 24
Public Const ucIsSorted = 25
Public Const ucIsMaybeComplex = 26
Public Const ucIsComplexAltDef = 27
Public Const ucIsArgTypeOverload = 28
Public Const ucIsSyntax = 29
Public Const Reserved1 = 30
Public Const Reserved2 = 31

Public Const ucInfixOp = ucIsInfixOp
Public Const ucPostfixOp = ucIsPostfixOp
Public Const ucPrefixOp = ucIsPrefixOp

' For ucElementCategory
Public Const ucIsOperand = 32
'ucIsFunction  = (already defined)
'ucIsInfixOp   = (already defined)
'ucIsPrefixOp  = (already defined)
'ucIsPostfixOp = (already defined)

Public Const ucDataType = 40
Public Const ucResultPtr = 41
Public Const ucSignPos = 42
Public Const ucElementCat = 43
Public Const ucTreeIndex = 44
Public Const ucErrorNumber = 45
Public Const ucOpHandle = 46
Public Const ucFuncHandle = ucOpHandle
Public Const ucExprPtr = 47
Public Const ucCodeAddress = 48
Public Const ucCurrentPtr = 49
Public Const ucParentPtr = 50
Public Const ucDataHandle = 51
Public Const ucMiscData = 52
Public Const ucUserData1 = 53
Public Const ucUserData2 = 54
Public Const ucUserData3 = 55
Public Const ucUserData4 = 56

Public Const ucTreePre = 5
Public Const ucTreeDownEarly = 6
Public Const ucTreeDown = 7
Public Const ucTreeUp = 8

Public Const ucErrorCallback = 9

Public Const ucNoChange = 0
Public Const ucMinorChange = 1
Public Const ucMajorChange = 2

' Trig mode options for ucSetTrigMode
Public Const ucRadians = 1
Public Const ucDegrees = 2
Public Const ucGrads = 3
Public Const ucCycles = 4

' Options for ucArrayInfo,
' and the first two for ucProperty.ArrayProperties
Public Const ucArray_BoundsCheck = 1
Public Const ucArray_ColumnMajor = 2
Public Const ucArray_DataAddress = 3
Public Const ucArray_DataType = 4
Public Const ucArray_DimensionCount = 5
Public Const ucArray_ElementCount = 6
Public Const ucArray_ElementSize = 7

' FPU control word values for ucProperty.FPU_ControlWord,
' and SetFPU(), SetOrFPU, SetXorFPU, GetFPU(), RestoreFPU
Public Const ucFPU_ErrMask_InvalidOp = 1
Public Const ucFPU_ErrMask_DenormalOp = 2
Public Const ucFPU_ErrMask_ZeroDivide = 4
Public Const ucFPU_ErrMask_Overflow = 8
Public Const ucFPU_ErrMask_Underflow = 16
Public Const ucFPU_ErrMask_PrecisionLoss = 32

Public Const ucFPU_Precision_Single = 0
Public Const ucFPU_Precision_Double = 512
Public Const ucFPU_Precision_Extended = 768

Public Const ucFPU_Round_Even = 0
Public Const ucFPU_Round_Down = 1024
Public Const ucFPU_Round_Up = 2048
Public Const ucFPU_Round_Truncate = 3072

' For t.ucProperty.FPU_Turbo
Public Const ucFPU_Safe = 0
Public Const ucFPU_SkipSet = 1
Public Const ucFPU_SkipReset = 65536

' For ucSetErrorData / ucGetErrorData
Public Const ucNewIndex = -2
Public Const ucNewUniqueIndex = -1
Public Const ucAllErrorData = -1

' Error Values
Public Const ucErr_None = 0
Public Const ucErr_ucErrorData = 1
Public Const ucErr_SyntaxError = 2
Public Const ucErr_MismatchedParenth = 3
Public Const ucErr_MismatchedQuotes = 4
Public Const ucErr_UndefinedVariable = 5
Public Const ucErr_UndefinedFunction = 6
Public Const ucErr_IncorrectArgType = 7
Public Const ucErr_InvalidNumberOfArgs = 8
Public Const ucErr_InvalidOperandType = 9
Public Const ucErr_InvalidVariableName = 10
Public Const ucErr_ImplicitMult = 11
Public Const ucErr_IncompleteBlock = 12

Public Const ucErr_InvalidOperation = 20
Public Const ucErr_DenormalOperand = 21
Public Const ucErr_DivisionByZero = 22
Public Const ucErr_Overflow = 23
Public Const ucErr_Underflow = 24
Public Const ucErr_InexactResult = 25
Public Const ucErr_SubscriptOutOfRange = 35
Public Const ucErr_StackOverflow = 36

Public Const ucErr_OperationDenied = 40
Public Const ucErr_PropertyConflict = 41
Public Const ucErr_ConstantWriteDenied = 42
Public Const ucErr_FunctionDefError = 43
Public Const ucErr_FunctionLocked = 44

Public Const ucErr_DefinitionSpaceFull = 60
Public Const ucErr_ThreadInitFailed = 61
Public Const ucErr_UnsupportedFeature = 62

Public Const Op_Func_GetFPU = 36
Public Const Op_Func_SetFPU = 37
Public Const Op_Func_SetOrFPU = 38
Public Const Op_Func_SetXorFPU = 39
Public Const Op_Func_RestoreFPU = 40
Public Const Op_Func_Ceil = 41
Public Const Op_Func_Factorial = 42
Public Const Op_Func_Not = 43
Public Const Op_Cur_Add = 44
Public Const Op_Cur_Multiply = 45
Public Const Op_Cur_Subtract = 46
Public Const Op_Cur_DivideToExt = 47
Public Const Op_Qwd_Add = 48
Public Const Op_Qwd_Multiply = 49
Public Const Op_Qwd_Subtract = 50
Public Const Op_Qwd_DivideToInt = 51
Public Const Op_Qwd_DivideToExt = 52
Public Const Op_Int_Add = 53
Public Const Op_Int_Multiply = 54
Public Const Op_Int_Subtract = 55
Public Const Op_Int_DivideToInt = 56
Public Const Op_Int_DivideToExt = 57
Public Const Op_Int_Mod = 58
Public Const Op_IntPower = 59

Public ucInitialized As Long
Public t0 As ucThreadType
Public ucItem() As ucHandleType

Public Declare Sub ucDetachSymbolE Lib "ucalcfmp.dll" (ByVal Symbol$, ByVal t As Long)
Public Declare Sub ucErrorClearE Lib "ucalcfmp.dll" (ByVal t As Long)
Public Declare Sub ucLicenseE Lib "ucalcfmp.dll" (ByVal SerialNumber$)
Public Declare Sub ucReleaseItemE Lib "ucalcfmp.dll" (ByVal ItemHandle As Long)
Public Declare Sub ucReleaseThreadE Lib "ucalcfmp.dll" (ByVal ucThread As Long)
Public Declare Sub ucResetE Lib "ucalcfmp.dll" (t As ucThreadType)
Public Declare Sub ucReturnDblE Lib "ucalcfmp.dll" (ByVal vValue As Double, t As ucThreadType)
Public Declare Sub ucReturnCplxDblE Lib "ucalcfmp.dll" (ByVal vValue As Double, t As ucThreadType)
Public Declare Sub ucReturnStrE Lib "ucalcfmp.dll" (ByVal Expression$, t As ucThreadType)
Public Declare Sub ucSetErrorDataE Lib "ucalcfmp.dll" (ByVal errIndex As Long, ByVal vValue As String, t As ucThreadType)
Public Declare Sub ucSetParamStrE Lib "ucalcfmp.dll" (ByVal ParamIndex As Long, ByVal Argument$, t As ucThreadType)
Public Declare Sub ucSetThreadAddrE Lib "ucalcfmp.dll" (ByVal ucThread As Long, t As ucThreadType)
Public Declare Sub ucSetTrigModeE Lib "ucalcfmp.dll" (ByVal TrigMode As Long, ByVal t As Long)
Public Declare Sub ucSetVariableValueDblE Lib "ucalcfmp.dll" (ByVal VarHandle As Long, ByVal vValue As Double)
Public Declare Sub ucSetVariableValueStrE Lib "ucalcfmp.dll" (ByVal VarHandle As Long, ByVal vValue As String)
Public Declare Sub ucTreeSetParamE Lib "ucalcfmp.dll" (ByVal pIndex As Long, ByVal pData$, t As ucThreadType)
Public Declare Function ucAliasE Lib "ucalcfmp.dll" (ByVal AssignFrom$, ByVal AssignTo$, ByVal aProperties As Long, ByVal Precedence As Long, ByVal t As Long) As Long
Public Declare Function ucArrayInfoE Lib "ucalcfmp.dll" (ByVal VarHandle As Long, ByVal InfoOption As Long) As Long
Public Declare Function ucDefineE Lib "ucalcfmp.dll" (ByVal Definition$, ByVal t As Long) As Long
Public Declare Function ucDefineFunctionE Lib "ucalcfmp.dll" (ByVal FunctionDef$, ByVal FunctionAddress As Long, ByVal Precedence As Long, ByVal fProperties As Long, ByVal vType As Long, ByVal t As Long) As Long
Public Declare Function ucDefineVariableE Lib "ucalcfmp.dll" (ByVal VariableDef$, ByVal vPtr As Long, ByVal Expr As Long, ByVal vProperties As Long, ByVal vType As Long, ByVal t As Long) As Long
Public Declare Function ucEvalE Lib "ucalcfmp.dll" (ByVal Expression$, ByVal t As Long) As Double
Public Declare Function ucEvalStrE Lib "ucalcfmp.dll" (ByVal Expression$, ByVal t As Long) As String
Public Declare Function ucEvaluateStrE Lib "ucalcfmp.dll" (ByVal ExpressionPtr As Long) As String
Public Declare Function ucGetErrorDataE Lib "ucalcfmp.dll" (ByVal errIndex As Long, ByVal t As Long) As String
Public Declare Function ucGetParamE Lib "ucalcfmp.dll" (ByVal pIndex As Long, t As ucThreadType) As Double
Public Declare Function ucGetParamCplxE Lib "ucalcfmp.dll" (ByVal pIndex As Long, t As ucThreadType) As Double
Public Declare Function ucGetParamStrE Lib "ucalcfmp.dll" (ByVal ParamIndex As Long, t As ucThreadType) As String
Public Declare Function ucGetThreadInfoE Lib "ucalcfmp.dll" (ByVal ucThread As Long, t As ucThreadType) As Long
Public Declare Function ucGetVariableValueE Lib "ucalcfmp.dll" (ByVal VarHandle As Long) As Double
Public Declare Function ucGetVariableValueStrE Lib "ucalcfmp.dll" (ByVal VarHandle As Long) As String
Public Declare Function ucImagPartE Lib "ucalcfmp.dll" (ByVal t As Long) As Double
Public Declare Function ucLowerBoundE Lib "ucalcfmp.dll" (ByVal VarHandle As Long, ByVal Dimension As Long) As Long
Public Declare Function ucNewThreadE Lib "ucalcfmp.dll" () As Long
Public Declare Function ucNewSymbolHandleE Lib "ucalcfmp.dll" (ByVal t As Long) As Long
Public Declare Function ucOpE Lib "ucalcfmp.dll" (ByVal OpID As Long) As Long
Public Declare Function ucParamHandleE Lib "ucalcfmp.dll" (ByVal pIndex As Long, t As ucThreadType) As Long
Public Declare Function ucParamTypeE Lib "ucalcfmp.dll" (ByVal pIndex As Long, t As ucThreadType) As Long
Public Declare Function ucParseE Lib "ucalcfmp.dll" (ByVal Expression$, ByVal LocalVars$, ByVal ExprType As Long, ByVal t As Long) As Long
Public Declare Function ucPrecedenceE Lib "ucalcfmp.dll" (ByVal Op$, ByVal OpCategory As Long, ByVal t As Long) As Long
Public Declare Function ucSetCallbackE Lib "ucalcfmp.dll" (ByVal FunctionAddress As Long, ByVal TreeOption As Long, ByVal t As Long) As Long
Public Declare Function ucSymbolHandleE Lib "ucalcfmp.dll" (ByVal ucSymbol$, ByVal SymbolProperty As Long, ByVal t As Long) As Long
Public Declare Function ucThreadAddrE Lib "ucalcfmp.dll" (ByVal t As Long) As Long
Public Declare Function ucTreeGetParamE Lib "ucalcfmp.dll" (ByVal pIndex As Long, t As ucThreadType) As String
Public Declare Function ucTreePropertyE Lib "ucalcfmp.dll" (ByVal NodePtr As Long, ByVal pIndex As Long, ByVal pOption As Long, t As ucThreadType) As Long
Public Declare Function ucUpperBoundE Lib "ucalcfmp.dll" (ByVal VarHandle As Long, ByVal Dimension As Long) As Long

' These 3 functions come from ucalcfmE.dll instead of ucalcfmP.dll
Public Declare Function ucEvaluate Lib "ucalcfme.dll" Alias "ucEvaluateE" (ByVal ExpressionPtr As Long) As Double
Public Declare Function ucEvaluateLng Lib "ucalcfme.dll" Alias "ucEvaluateLngE" (ByVal ExpressionPtr As Long) As Long
Public Declare Function ucEvaluateFxt Lib "ucalcfme.dll" Alias "ucEvaluateFxtE" (ByVal ExpressionPtr As Long) As Currency
Public Declare Sub ucEvaluateInitE Lib "ucalcfme.dll" ()
Public Declare Sub ucReleaseExprPtrE Lib "ucalcfme.dll" (ByVal ExpressionPtr As Long)

Public Function ucOp(ByVal OpID As Long) As Long
   ucOp = ucOpE(OpID)
End Function

Public Function ucPrecedence(Op$, Optional ByVal OpCategory As Long, Optional ByVal t As Long = 0) As Long
   ucPrecedence = ucPrecedenceE(Op$, OpCategory, ucThreadAddr(t))
End Function

Public Function ucSymbolHandle(ucSymbolE$, Optional ByVal SymbolProperty As Long = 0, Optional ByVal t As Long = 0) As Long
   ucSymbolHandle = ucSymbolHandleE(ucSymbolE$, SymbolProperty, ucThreadAddr(t))
End Function

Public Function ucError() As Long
   ucError = t0.ErrorNumber
End Function

Public Property Get ucErrorData(ByVal errIndex As Long, Optional ByVal t As Long = 0) As String
   ucErrorData = ucGetErrorDataE(errIndex, ucThreadAddr(t))
End Property

Public Property Get ucGetErrorData(ByVal errIndex As Long, Optional ByVal t As Long = 0) As String
   ucGetErrorData = ucGetErrorDataE(errIndex, ucThreadAddr(t))
End Property

Public Function ucAttachVariable(VariableDef$, ByRef vPtr As Long, Optional ByVal ExprPtr As Long = 0, Optional ByVal vProperties As Long = 0, Optional ByVal t As Long = 0) As Long
   ucAttachVariable = ucDefineVariableE(VariableDef$, vPtr, ExprPtr, vProperties, t0.ucProperty.DefaultNumericType, ucThreadAddr(t))
End Function

Public Function ucDefine(Definition$) As Long
   ucDefine = ucDefineE(Definition$, ucThreadAddr(0))
End Function

Public Function ucDefineFunction(FunctionDef$, Optional ByVal FunctionAddress As Long = 0, Optional ByVal fProperties As Long = 0, Optional ByVal vType As Long = 0)
   ucDefineFunction = ucDefineFunctionE(FunctionDef$, FunctionAddress, 0, fProperties, vType, ucThreadAddr(0))
End Function

Public Function ucDefineMacro(Definition$) As Long
   ucDefineMacro = ucDefineE("Macro " + Definition$, ucThreadAddr(0))
End Function

Public Function ucDefineOperator(OperatorDef$, Optional ByVal Precedence As Long = 0, Optional ByVal OperatorAddress As Long = 0, Optional ByVal fProperties As Long = 0, Optional ByVal vType As Long = 0)
   ucDefineOperator = ucDefineFunctionE(OperatorDef$, OperatorAddress, Precedence, fProperties, vType, ucThreadAddr(0))
End Function

Public Function ucDefineSyntax(Definition$) As Long
   ucDefineSyntax = ucDefineE("Syntax " + Definition$, ucThreadAddr(0))
End Function

Public Function ucAlias(AssignFrom$, AssignTo$, Optional ByVal aProperties As Long = 0, Optional ByVal Precedence As Long = 0) As Long
   ucAlias = ucAliasE(AssignFrom$, AssignTo$, aProperties, Precedence, ucThreadAddr(0))
End Function

Public Sub ucDetachSymbol(Symbol$, ByVal t As Long)
   ucDetachSymbolE Symbol$, ucThreadAddr(t)
End Sub

Public Sub ucErrorClear()
   ucErrorClearE ucThreadAddr(0)
End Sub

Public Function ucDefineConstant(ConstantDef$, Optional ByVal vProperties As Long = 0) As Long
   ucDefineConstant = ucDefineVariableE(ConstantDef$, 0, 0, vProperties + ucConstant, t0.ucProperty.DefaultNumericType, ucThreadAddr(0))
End Function

Public Sub ucReset(t As ucThreadType)
   Dim Between_Mult_Pow As Long
   Dim Between_And_Less As Long
   Dim Between_Div_Mod As Long
   
   ReDim ucItem(t.ucProperty.DefSpace.BaseDefinitionSpace _
               + t.ucProperty.DefSpace.DefinitionSpace _
               * t.ucProperty.DefSpace.MaxThreadCount) As ucHandleType
   
   t.Context = t.Context Or ucContext_ucReset
   t.Handshake.LanguageID = uc_Lang_VisualBasic
   t.Handshake.HandleAddr = VarPtr(ucItem(0).ucThread)
   ucEvaluateInitE
   ucResetE t
   
   t.ucProperty.ComplexNumbers = ucTrue
   t.ucProperty.EscapeChar = Asc("~")
   t.ucProperty.DefaultExternalStringType = ucType_StringBSTR
   
   ucDefineConstant "Byte     = uc_Integer8u"
   ucDefineConstant "Integer  = uc_Integer16"
   ucDefineConstant "Word     = uc_Integer16u"
   ucDefineConstant "Long     = uc_Integer32"
   ucDefineConstant "DWord    = uc_Integer32u"
   ucDefineConstant "Quad     = uc_Integer64"
   ucDefineConstant "Currency = uc_Currency"
   ucDefineConstant "CurrencyX= uc_CurrencyExt"
   ucDefineConstant "Single   = uc_FloatSingle"
   ucDefineConstant "Double   = uc_FloatDouble"
   ucDefineConstant "Extended = uc_FloatExtended"
   ucDefineConstant "AsciiZ   = uc_StringLPSTR"
   ucDefineConstant "String   = uc_StringAnsi"
   ucDefineConstant "FixedPt  = Currency"
   ucDefineConstant "Any      = uc_Any"
   
   Between_Mult_Pow = (ucPrecedence("*") + ucPrecedence("^")) / 2
   Between_And_Less = (ucPrecedence("AND") + ucPrecedence("<")) / 2
   Between_Div_Mod = (ucPrecedence("/") + ucPrecedence("MOD")) / 2
   
   t.ucProperty.FunctionProperties = ucFastCallback + ucOverload

   ucDefineOperator "*[x As Long, y As Long] As Long", ucPrecedence("*"), ucOp(Op_Int_Multiply)
   ucDefineOperator "-[x As Long, y As Long] As Long", ucPrecedence("-"), ucOp(Op_Int_Subtract)
   ucDefineOperator "+[x As Long, y As Long] As Long", ucPrecedence("+"), ucOp(Op_Int_Add)
   ucDefineOperator "/[x As Long, y As Long] As Extended", ucPrecedence("/"), ucOp(Op_Int_DivideToExt)
   ucDefineOperator "\[x As Long, y As Long] As Long", Between_Div_Mod, ucOp(Op_Int_DivideToInt)
   ucDefineOperator "Mod[x As Long, y As Long] As Long", ucPrecedence("MOD"), ucOp(Op_Int_Mod)
   ucDefineOperator "^[x As Extended, y As Long] As Extended", ucPrecedence("^"), ucOp(Op_IntPower)

   ucDefineOperator "*[x As Quad, y As Quad] As Quad", ucPrecedence("*"), ucOp(Op_Qwd_Multiply)
   ucDefineOperator "-[x As Quad, y As Quad] As Quad", ucPrecedence("-"), ucOp(Op_Qwd_Subtract)
   ucDefineOperator "+[x As Quad, y As Quad] As Quad", ucPrecedence("+"), ucOp(Op_Qwd_Add)
   ucDefineOperator "/[x As Quad, y As Quad] As Extended", ucPrecedence("/"), ucOp(Op_Qwd_DivideToExt)
   ucDefineOperator "\[x As Quad, y As Quad] As Quad", Between_Div_Mod, ucOp(Op_Qwd_DivideToInt)

   ucDefineOperator "*[x As FixedPt, y As FixedPt] As FixedPt", ucPrecedence("*"), ucOp(Op_Cur_Multiply)
   ucDefineOperator "-[x As FixedPt, y As FixedPt] As FixedPt", ucPrecedence("-"), ucOp(Op_Cur_Subtract)
   ucDefineOperator "+[x As FixedPt, y As FixedPt] As FixedPt", ucPrecedence("+"), ucOp(Op_Cur_Add)
   ucDefineOperator "/[x As FixedPt, y As FixedPt] As Extended", ucPrecedence("/"), ucOp(Op_Cur_DivideToExt)

   ucDefineFunction "Fact(x As Long)", ucOp(Op_Func_Factorial)
   ucDefineFunction "GetFPU(x As Long) As Long", ucOp(Op_Func_GetFPU)
   ucDefineFunction "SetFPU(x As Long)", ucOp(Op_Func_SetFPU)
   ucDefineFunction "SetOrFPU(x As Long)", ucOp(Op_Func_SetOrFPU)
   ucDefineFunction "SetXorFPU(x As Long)", ucOp(Op_Func_SetXorFPU)
   ucDefineFunction "RestoreFPU(x As Long)", ucOp(Op_Func_RestoreFPU)
   ucDefineFunction "Ceil(x)", ucOp(Op_Func_Ceil)
   ucDefineFunction "Not(x)", ucOp(Op_Func_Not)

   ' String relational (comparison) operators
   t.ucProperty.DefaultPrecedence = ucPrecedence("<")
   t.ucProperty.FunctionProperties = ucDynamic + ucLocked + ucOverride + ucOverload
   ucDefineFunction "StrCmp(a As String, b As String) = StrCmp(0, 0, a, b)"
   ucDefineOperator " =[a As String, b As String] = StrCmp(a, b) = 2"
   ucDefineOperator "<>[a As String, b As String] = StrCmp(a, b) <> 2"
   ucDefineOperator " >[a As String, b As String] = StrCmp(a, b) = 3"
   ucDefineOperator ">=[a As String, b As String] = StrCmp(a, b) > 1"
   ucDefineOperator " <[a As String, b As String] = StrCmp(a, b) = 1"
   ucDefineOperator "<=[a As String, b As String] = StrCmp(a, b) < 3"
   
   t.ucProperty.FunctionProperties = ucInline + ucLocked + ucOverride + ucOverload
   
   ucDefineFunction "Precedence(Operator$, OpCategory = 0)", AddressOf uc_Precedence
   ucDefineFunction "PrecedenceMax()", AddressOf uc_PrecedenceMax
   ucDefineFunction "PrecedenceMin()", AddressOf uc_PrecedenceMin
   
   ucDefineMacro ".[x]  ::= (0.x) ::: Precedence = PrecedenceMax"
   ucDefineMacro " [x]. ::= (x.0) ::: Precedence = PrecedenceMax"

   ' Hex, binary, octal notations.  Includes support for non-integers
   ucDefineSyntax "#b (Number:'[0-1]+') [(Frac:'\.[0-1]+'=)] ::= BaseConvert(~qNumber~jFrac~q, 2) ::: Precedence = PrecedenceMax"
   ucDefineSyntax "#o (Number:'[0-7]+') [(Frac:'\.[0-7]+'=)] ::= BaseConvert(~qNumber~jFrac~q, 8) ::: Precedence = PrecedenceMax"
   ucDefineSyntax "#h (Number:'[0-9A-F]+') [(Frac:'\.[0-9A-F]+'=)] ::= BaseConvert(~qNumber~jFrac~q, 16) ::: Precedence = PrecedenceMax"
   ucDefineSyntax "#i ::= ~i"
  

   ' Alternative notations
'   ucDefineOperator "  &[a,b] = a and b", ucPrecedence("and")
'   ucDefineOperator "  |[a,b] = a or  b", ucPrecedence("or ")
'   ucDefineOperator " =<[a,b] = a <=  b", ucPrecedence("<= ")
'   ucDefineOperator " =>[a,b] = a >=  b", ucPrecedence(">= ")
   ucDefineOperator " ==[a,b] = a  =  b", ucPrecedence("=")
   ucDefineOperator "DIV[a,b] = a  \  b", ucPrecedence("\")
   ucDefineOperator "**[a,b]  = a  ^  b", ucPrecedence("^")
   ucDefineMacro    " &[TextA,TextB] ::= TextA + TextB"
   ucDefineOperator "==[TextA$,TextB$] = TextA = TextB", ucPrecedence("=")
   ucDefineOperator "Imp[a, b] = (Not(a)) Or b", ucPrecedence("OR")-1
   ucDefineOperator "Eqv[a, b] = (a and b) Or (not(a) and not(b))", ucPrecedence("OR")-1
   ucDefineFunction "Concatenate$(a$,b$) = a + b"
   ucDefineFunction "Sign(x) = Sgn(x)"
   ucDefineFunction "Sqrt(x) = Sqr(x)"
   ucDefineFunction "Ln(x)   = Log(x)"
   ucDefineFunction "Atan(x) = Atn(x)"
   ucDefineOperator "[x]! = fact(x)", t.PrecedenceMax
   ucDefineOperator "[x]% = x/100", t.PrecedenceMax
   ucDefineOperator "%[x,y] = x mod y", ucPrecedence("MOD")
   ucDefineOperator "-[x] = 0-x", Between_Mult_Pow
   ucDefineOperator "+[x] = 0+x", Between_Mult_Pow
   ucDefineOperator "Not[x] = Not(x)", Between_And_Less

   ' FPU routines
   ucDefineFunction "GetFPU() = GetFPU(0)"
   ucDefineFunction "RestoreFPU() = RestoreFPU(0)"
   ucAlias "GetFPU", "GetFPU(0)"
   ucAlias "RestoreFPU", "RestoreFPU(0)"
   ucAlias "SetFPU(setting, Expression)", "(SetFPU(setting)+(Expression)+RestoreFPU(0))"
   ucAlias "SetOrFPU(setting, Expression)", "(SetOrFPU(setting)+(Expression)+RestoreFPU(0))"
   ucAlias "SetXorFPU(setting, Expression)", "(SetXorFPU(setting)+(Expression)+RestoreFPU(0))"

   ucDefineFunction "Inf() = SetOrFPU(4, 1/0)", 0, ucDynamic
   ucDefineFunction "NaN() = SetOrFPU(1, 0/0)", 0, ucDynamic
   
   ' Hyperbolic functions
   ucDefineFunction "Sinh(x)  = (exp(x) - exp(-x)) / 2"
   ucDefineFunction "Cosh(x)  = (exp(x) + exp(-x)) / 2"
   ucDefineFunction "Tanh(x)  = (exp(2x) - 1) / (exp(2x) + 1)"
   ucDefineFunction "Coth(x)  = (exp(2x) + 1) / (exp(2x) - 1)"
   ucDefineFunction "Sech(x)  = 2 / (exp(x) + exp(-x))"
   ucDefineFunction "Csch(x)  = 2 / (exp(x) - exp(-x))"
   ucDefineFunction "Asinh(x) = Log(x + Sqr(x ^ 2 + 1))"
   ucDefineFunction "Acosh(x) = Log(x + Sqr(x ^ 2 - 1))"
   ucDefineFunction "Atanh(x) = Log((1 + x) / (1 - x)) / 2"
   ucDefineFunction "Acoth(x) = Log((x + 1) / (x - 1)) / 2"
   ucDefineFunction "Asech(x) = Log((1 + Sqr(1 - x ^ 2)) / x)"
   ucDefineFunction "Acsch(x) = Log(1 / x + Sqr(1 + x ^ 2) / Abs(x))"

   ' A few constants needed in some definitions further down
   ucDefineConstant "Pi     = Atn(1) * 4"
   ucDefineConstant "HalfPI = Pi/2"

   ' Other functions
   ucDefineFunction "Sec(x)   = 1 / Cos(x)"
   ucDefineFunction "Csc(x)   = 1 / Sin(x)"
   ucDefineFunction "Cot(x)   = 1 / Tan(x)"
   ucDefineFunction "Asin(x)  = SetOrFPU(4, Atn(x / Sqr(1 - x * x)))" ' +++ make provisions for complex
   ucDefineFunction "Acos(x)  = HalfPI - Asin(x)"
   ucDefineFunction "Log2(x)  = Log(x) / Log(2)"
   ucDefineFunction "Exp2(x)  = 2 ^ x"
   ucDefineFunction "Log10(x) = Log(x) / Log(10)"
   ucDefineFunction "Exp10(x) = 10 ^ x"

   ' Complex operators & functions
   t.ucProperty.FunctionProperties = ucDynamic + ucLocked + ucOverride + ucOverload + ucComplex
   ucDefineOperator "+[x] = x.Real + x.Imag*#i", Between_Mult_Pow
   ucDefineOperator "-[x] = (-x.Real) + (-x.Imag)*#i", Between_Mult_Pow
   
   ucDefineOperator "+[x,y] = x.Real+y.Real + (x.Imag+y.Imag)*#i", ucPrecedence("+")
   ucDefineOperator "*[x,y] = (x.Real*y.Real - x.Imag*y.Imag)+(x.Real*y.Imag + x.Imag*y.Real)*#i", ucPrecedence("*")
   ucDefineOperator "-[x,y] = (x.Real-y.Real) + (x.Imag-y.Imag)*#i", ucPrecedence("-")
   ucDefineOperator "/[x,y] = x.Real*y.Real/(y.Real^2+y.Imag^2) + x.Imag*y.Imag" _
   + "/(y.Real^2+y.Imag^2) + (x.Imag*y.Real/(y.Real^2+y.Imag^2)-x.Real*y.Imag" _
   + "/(y.Real^2+y.Imag^2))*#i", ucPrecedence("/")
  
   ucDefineFunction "Abs(x) = Sqr(x.Real^2 + x.Imag^2) + 0*#i"
   ucDefineFunction "Sgn(x) = x.Real/Abs(x) + x.Imag/Abs(x)*#i"
   ucDefineFunction "Sin(x) = Sin(x.Real)*Cosh(x.Imag) + Cos(x.Real)*Sinh(x.Imag)*#i"
   ucDefineFunction "Cos(x) = Cos(x.Real)*Cosh(x.Imag) - Sin(x.Real)*Sinh(x.Imag)*#i"
   ucDefineFunction "Phase(x) = iif(x.Imag=0, iif(x.Real>0, 0, pi), atn(x.Imag/x.Real)) + 0*#i"
   ucDefineFunction "Conj(x)  = x.Real - x.Imag*#i"
   ucDefineFunction "Exp(x)   = exp(x.Real)*Cos(x.Imag) + exp(x.Real)*sin(x.Imag)*#i"
   ucDefineFunction "Imag(x)  = x.Imag + 0*#i"
   ucDefineFunction "Real(x)  = x.Real + 0*#i"
   
   ucDefineFunction "Sqr(x)   = Sqr(Sqr(x.Real^2 + x.Imag^2)+x.Real)/Sqr(2)" _
   + "+ (Sqr(Sqr(x.Real^2 + x.Imag^2)-x.Real)*iif(x.Imag < 0, -1, 1)/Sqr(2))*#i"
   ucDefineFunction "Sqrt(x)  = Sqr(x.Real+x.Imag*#i)"

   ucDefineFunction "Log(x) = iif(x.Real=0 and x.Imag=0, -Inf," _
   + "Log(x.Real^2 + x.Imag^2)/2)" _
   + "+iif(x.Real=0 and x.Imag=0, 0, iif(x.Real>0 and x.Imag=0, 0," _
   + "iif(x.Imag=0, pi, pi*sgn(x.Imag)/2-Atn(x.Real/x.Imag))))*#i"

   ucDefineFunction "_ZeroToZero() = 1", 0, ucDynamic + ucOverwrite
'  ucDefineFunction "_ZeroToZero() = 0", 0, ucDynamic+ucOverwrite
'  ucDefineFunction "_ZeroToZero() = NaN", 0, ucDynamic+ucOverwrite
'  ucDefineFunction "_ZeroToZero() = SetError(1, '0^0 Error')", 0, ucDynamic+ucOverwrite

   ucDefineOperator "^[x,y] = iif(x.Imag = 0, iif(x.Real > 0," _
   + "x.Real^y.Real*Cos(y.Imag*Log(x.Real))," _
   + "iif(x.Real=0 and y.Real=0 and y.Imag=0, _ZeroToZero()," _
   + "(-x.Real)^y.Real*COS(y.Imag*LOG(-x.Real)+pi*y.Real)/EXP(pi*y.Imag)))," _
   + "EXP(y.Imag*ATN(x.Real/x.Imag))*SQR((x.Imag^2+x.Real^2)^y.Real)" _
   + "*COS(pi*y.Real*sgn(x.Imag)/2-y.Real*ATN(x.Real/x.Imag)+y.Imag" _
   + "*LOG(x.Imag^2+x.Real^2)/2)/SQR(EXP(pi*y.Imag*Sgn(x.Imag))))-55+55" _
   + "+iif(x.Imag = 0, iif(x.Real > 0," _
   + "x.Real^y.Real*SIN(y.Imag*LOG(x.Real))," _
   + "iif(x.Real=0 and y.Real=0 and y.Imag=0, 0," _
   + "(-x.Real)^y.Real*SIN(y.Imag*LOG(-x.Real)+pi*y.Real)/EXP(pi*y.Imag)))," _
   + "EXP(y.Imag*ATN(x.Real/x.Imag))*SQR((x.Imag^2+x.Real^2)^y.Real)" _
   + "*SIN(pi*y.Real*sgn(x.Imag)/2-y.Real*ATN(x.Real/x.Imag)+y.Imag" _
   + "*LOG(x.Imag^2+x.Real^2)/2)/SQR(EXP(pi*y.Imag*Sgn(x.Imag)))-55+55)*#i", _
     ucPrecedence("^")
     
   t.ucProperty.FunctionProperties = ucDynamic + ucOverload + ucOverwrite + ucComplex + ucRecursion
   t.ucProperty.DefaultNumericType = ucType_Double
   t.Context = t.Context Xor ucContext_ucReset
   t.ucProperty.DefaultPrecedence = Between_Div_Mod ' arbitrarily chosen
End Sub

Public Function ucGetThreadInfo(ByVal ucThread As Long, t As ucThreadType) As Long
   ucGetThreadInfo = ucGetThreadInfoE(ucThread, t)
End Function

Public Sub ucSetThreadAddr(ByVal ucThread As Long, t As ucThreadType)
   ucSetThreadAddrE ucThread, t
End Sub

Public Sub ucSetTrigMode(ByVal TrigMode As Long, Optional ByVal t As Long)
   ucSetTrigModeE TrigMode, ucThreadAddr(t)
End Sub

Public Function ucArrayInfo(ByVal VarHandle As Long, ByVal InfoOption As Long) As Long
   ucArrayInfo = ucArrayInfoE(VarHandle, InfoOption)
End Function

Public Function ucLowerBound(ByVal VarHandle As Long, Optional ByVal Dimension As Long = 1) As Long
   ucLowerBound = ucLowerBoundE(VarHandle, Dimension)
End Function

Public Function ucParse(Expression$, Optional LocalVars$ = "", Optional ExprType As Long = 0) As Long
   ucParse = ucParseE(Expression$, LocalVars$, ExprType, ucThreadAddr(0))
End Function

Public Function ucUpperBound(ByVal VarHandle As Long, Optional ByVal Dimension As Long = 1) As Long
   ucUpperBound = ucUpperBoundE(VarHandle, Dimension)
End Function

Public Function ucNewSymbolHandle(Optional ByVal t As Long = 0) As Long
   ucNewSymbolHandle = ucNewSymbolHandleE(ucThreadAddr(t))
End Function

Public Sub ucReturn(ByVal vValue As Double, t As ucThreadType)
   ucReturnDblE vValue, t
End Sub

Public Sub ucReturnCplx(ByVal vValue As Double, t As ucThreadType)
   ucReturnCplxDblE vValue, t
End Sub

Public Sub ucReturnStr(Expression$, t As ucThreadType)
   ucReturnStrE Expression$, t
End Sub

Public Function ucErrorHandler(ByVal FunctionAddress As Long) As Long
   ucErrorHandler = ucSetCallbackE(FunctionAddress, ucErrorCallback, ucThreadAddr(0))
End Function

Public Function ucEval(Expression$, Optional ByVal t As Long = 0) As Double
   ucEval = ucEvalE(Expression$, ucThreadAddr(t))
End Function

Public Function ucEvalStr(Expression$, Optional ByVal t As Long = 0) As String
   ucEvalStr = ucEvalStrE(Expression$, ucThreadAddr(t))
End Function

Public Function ucEvaluateStr(ByVal ExpressionPtr As Long) As String
   ucEvaluateStr = ucEvaluateStrE(ExpressionPtr)
End Function

Public Function ucGetVariableValue(ByVal VarHandle As Long) As Double
   ucGetVariableValue = ucGetVariableValueE(VarHandle)
End Function

Public Function ucGetVariableValueStr(ByVal VarHandle As Long) As String
   ucGetVariableValueStr = ucGetVariableValueStrE(VarHandle)
End Function

Public Property Get ucVariableValue(ByVal VarHandle As Long) As Double
   ucVariableValue = ucGetVariableValueE(VarHandle)
End Property

Public Property Let ucVariableValue(ByVal VarHandle As Long, ByVal vValue As Double)
   ucSetVariableValueDblE VarHandle, vValue
End Property

Public Property Get ucVariableValueStr(ByVal VarHandle As Long) As String
   ucVariableValueStr = ucGetVariableValueStrE(VarHandle)
End Property

Public Property Let ucVariableValueStr(ByVal VarHandle As Long, vValue As String)
   ucSetVariableValueStrE VarHandle, vValue
End Property

Public Sub ucSetVariableValue(ByVal VarHandle As Long, ByVal vValue As Double)
   ucSetVariableValueDblE VarHandle, vValue
End Sub

Public Sub ucSetVariableValueStr(ByVal VarHandle As Long, vValue$)
   ucSetVariableValueStrE VarHandle, vValue$ + Chr$(0)
End Sub

Public Function ucImagPart(Optional ByVal t As Long = 0) As Double
   ucImagPart = ucImagPartE(ucThreadAddr(t))
End Function

Public Function ucPreParse(ByVal FunctionAddress As Long) As Long
   ucPreParse = ucSetCallbackE(FunctionAddress, ucTreePre, ucThreadAddr(0))
End Function

Public Function ucTreeParseDownEarly(ByVal FunctionAddress As Long) As Long
   ucTreeParseDownEarly = ucSetCallbackE(FunctionAddress, ucTreeDownEarly, ucThreadAddr(0))
End Function

Public Function ucTreeParseDown(ByVal FunctionAddress As Long) As Long
   ucTreeParseDown = ucSetCallbackE(FunctionAddress, ucTreeDown, ucThreadAddr(0))
End Function

Public Function ucTreeParseUp(ByVal FunctionAddress As Long) As Long
   ucTreeParseUp = ucSetCallbackE(FunctionAddress, ucTreeUp, ucThreadAddr(0))
End Function

Public Function ucGetExpr(t As ucThreadType) As String
   ucGetExpr = ucTreeGetParamE(-1, t)
End Function

Public Function ucTreeGetExpr(t As ucThreadType) As String
   ucTreeGetExpr = ucTreeGetParamE(-1, t)
End Function

Public Function ucTreeGetFunc(t As ucThreadType) As String
   ucTreeGetFunc = ucTreeGetParamE(0, t)
End Function

Public Function ucTreeGetOp(t As ucThreadType) As String
   ucTreeGetOp = ucTreeGetParamE(0, t)
End Function

Public Function ucTreeGetParam(ByVal pIndex As Long, t As ucThreadType) As String
   ucTreeGetParam = ucTreeGetParamE(pIndex, t)
End Function

Public Function ucTreeGetParamList(t As ucThreadType) As String
   ucTreeGetParamList = ucTreeGetParamE(-2, t)
End Function

Public Sub ucSetExpr(Expression As String, t As ucThreadType)
   ucTreeSetParamE -1, Expression, t
End Sub

Public Sub ucTreeSetExpr(Expression As String, t As ucThreadType)
   ucTreeSetParamE -1, Expression, t
End Sub

Public Sub ucTreeSetFunc(TreeFunc As String, t As ucThreadType)
   ucTreeSetParamE 0, TreeFunc, t
End Sub

Public Sub ucTreeSetOp(TreeOp As String, t As ucThreadType)
   ucTreeSetParamE 0, TreeOp, t
End Sub

Public Sub ucTreeSetParam(ByVal pIndex As Long, pData As String, t As ucThreadType)
   ucTreeSetParamE pIndex, pData, t
End Sub

Public Sub ucTreeSetParamList(pData As String, t As ucThreadType)
   ucTreeSetParamE -2, pData, t
End Sub

Public Function ucElementCategory(t As ucThreadType) As Long
   ucElementCategory = ucTreePropertyE(0, 0, ucElementCat, t)
End Function

Public Function ucTreeProperty(ByVal NodePtr As Long, ByVal pIndex As Long, ByVal pOption As Long, t As ucThreadType) As Long
   ucTreeProperty = ucTreePropertyE(NodePtr, pIndex, pOption, t)
End Function

Public Function ucTreeNodeProperty(ByVal pOption As Long, t As ucThreadType) As Long
   ucTreeNodeProperty = ucTreePropertyE(0, 0, pOption, t)
End Function

Public Function ucTreeParamProperty(ByVal pIndex As Long, ByVal pOption As Long, t As ucThreadType) As Long
   ucTreeParamProperty = ucTreePropertyE(0, pIndex, pOption, t)
End Function

Public Function ucTreeParentProperty(ByVal pOption As Long, t As ucThreadType) As Long
   Dim ParentPtr As Long

   ParentPtr = ucTreePropertyE(0, 0, ucParentPtr, t)
   ucTreeParentProperty = ucTreePropertyE(ParentPtr, 0, pOption, t)
End Function

Public Function ucDefineVariable(VariableDef$, Optional ByVal vProperties As Long = 0) As Long
   ucDefineVariable = ucDefineVariableE(VariableDef$, 0, 0, vProperties, t0.ucProperty.DefaultNumericType, ucThreadAddr(0))
End Function

Public Function ucGetParamStr(ByVal ParamIndex As Long, t As ucThreadType) As String
   ucGetParamStr = ucGetParamStrE(ParamIndex, t)
End Function
Public Sub ucSetParamStr(ByVal ParamIndex As Long, Argument$, t As ucThreadType)
   ucSetParamStrE ParamIndex, Argument$, t
End Sub

Public Property Get ucParamStr(ByVal ParamIndex As Long, t As ucThreadType) As String
   ucParamStr = ucGetParamStrE(ParamIndex, t)
End Property
Public Property Let ucParamStr(ByVal ParamIndex As Long, t As ucThreadType, ByVal Argument As String)
   ucSetParamStrE ParamIndex, Argument, t
End Property

Public Sub ucReleaseExprPtr(ByVal ExpressionPtr As Long)
   ucReleaseExprPtrE ExpressionPtr
End Sub

Public Sub ucReleaseItem(ByVal ItemHandle As Long)
   ucReleaseItemE ItemHandle
End Sub

'Public Property Let ucErrorData(ByVal errIndex As Long, t As ucThreadType, vValue As String)
'   ucSetErrorDataE errIndex, vValue, t
'End Property ' +++

Public Sub ucLicense(SerialNumber$)
   ucLicenseE SerialNumber$
End Sub

Public Function ucParam(ByVal x As Long, t As ucThreadType) As Double
   ucParam = ucGetParamE(x, t)
End Function

Public Function ucParamCplx(ByVal x As Long, t As ucThreadType) As Double
   ucParamCplx = ucGetParamCplxE(x, t)
End Function

Public Function ucParamHandle(ByVal pIndex As Long, t As ucThreadType) As Long
   ucParamHandle = ucParamHandleE(pIndex, t)
End Function

Public Function ucParamType(ByVal pIndex As Long, t As ucThreadType) As Long
   ucParamType = ucParamTypeE(pIndex, t)
End Function

Public Sub ucSetErrorData(ByVal errIndex As Long, vValue As String, t As ucThreadType)
   ucSetErrorDataE errIndex, vValue, t
End Sub

Private Function uc_Precedence(t As ucThreadType) As Double
   uc_Precedence = ucPrecedence(ucParamStr(1, t), ucParam(2, t), t.ucThread)
End Function

Private Function uc_PrecedenceMax(t As ucThreadType) As Double
   uc_PrecedenceMax = t.PrecedenceMax
End Function

Private Function uc_PrecedenceMin(t As ucThreadType) As Double
   uc_PrecedenceMin = t.PrecedenceMin
End Function

Public Function ucEvaluateCplx(ByVal ExpressionPtr As Long, ucValue As ucComplexNumberType, Optional ByVal t As Long = 0) As Double
   ucValue.Real = ucEvaluate(ExpressionPtr)
   ucValue.Imag = ucImagPartE(ucThreadAddr(t))
   ucEvaluateCplx = ucValue.Real
End Function

#If ucAltErrMsg = 0 Then
Public Function ucGetErrorMessage(Optional ByVal ErrorNumber As Long = -1, Optional ByVal t As Long = 0) As String
   Dim ErrMsg As String

   If ErrorNumber = -1 Then ErrorNumber = t0.ErrorNumber

   Select Case ErrorNumber
   Case ucErr_None
      ' No Error
   Case ucErr_ucErrorData
      ErrMsg = ucErrorData(0, t)
   Case ucErr_SyntaxError
      ErrMsg = "Syntax error"
   Case ucErr_MismatchedParenth
      ErrMsg = "Mismatched parenthesis"
   Case ucErr_MismatchedQuotes
      ErrMsg = "Mismatched quotes"
   Case ucErr_UndefinedVariable
      ErrMsg = "Undefined variable"
   Case ucErr_UndefinedFunction
      ErrMsg = "Undefined function"
   Case ucErr_IncorrectArgType
      ErrMsg = "Incorrect argument type"
   Case ucErr_InvalidNumberOfArgs
      ErrMsg = "Invalid number of function arguments"
   Case ucErr_InvalidOperandType
      ErrMsg = "Invalid operand type"
   Case ucErr_InvalidVariableName
      ErrMsg = "Invalid variable name"
   Case ucErr_ImplicitMult
      ErrMsg = "Implicit multiplication error"
   Case ucErr_IncompleteBlock
      ErrMsg = "Incomplete syntax block"

   Case ucErr_InvalidOperation
      ErrMsg = "Invalid floating point operation"
   Case ucErr_DenormalOperand
      ErrMsg = "Denormal floating point operand"
   Case ucErr_DivisionByZero
      ErrMsg = "Division by 0"
   Case ucErr_Overflow
      ErrMsg = "Floating point overflow"
   Case ucErr_Underflow
      ErrMsg = "Floating point underflow"
   Case ucErr_InexactResult
      ErrMsg = "Inexact floating point result"
   Case ucErr_SubscriptOutOfRange
      ErrMsg = "Array subscript out of range"
   Case ucErr_StackOverflow
      ErrMsg = "Stack overflow"

   Case ucErr_OperationDenied
      ErrMsg = "Cannot perform requested operation"
   Case ucErr_PropertyConflict
      ErrMsg = "Definition property conflict"
   Case ucErr_ConstantWriteDenied
      ErrMsg = "Assignment to a constant not allowed"
   Case ucErr_FunctionDefError
      ErrMsg = "Invalid function or operator definition"
   Case ucErr_FunctionLocked
      ErrMsg = "Function or operator is locked"

   Case ucErr_DefinitionSpaceFull
      ErrMsg = "Definition space is full"
   Case ucErr_ThreadInitFailed
      ErrMsg = "Thread initialization failed"
   Case ucErr_UnsupportedFeature
      ErrMsg = "Requires advanced edition of UCalc FMP"
   Case Else
      ErrMsg = "Unknown error"
      ' numbers up to 500 are reserved for future use by UCalc
      ' Use Case 501 and beyond for user-defined errors
   End Select

   If ErrorNumber > 1 Then ErrMsg = "Error #" + LTrim$(Str$(ErrorNumber)) + ": " + ErrMsg
   ucGetErrorMessage = ErrMsg
End Function

Public Function ucErrorMessage(Optional ByVal ErrorNumber As Long = -1, Optional ByVal t As Long = 0) As String
   ucErrorMessage = ucGetErrorMessage(ErrorNumber, t)
End Function

#End If

Public Function ucThreadAddr(ByVal t As Long) As Long
   If ucInitialized = False Then Call ucEvaluateInit
   ucThreadAddr = ucThreadAddrE(t)
End Function

Public Sub ucEvaluateInit()
   ucInitialized = True
   ucGetThreadInfo 0, t0
   ucSetThreadAddr 0, t0
   ucReset t0
End Sub
